# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Utilities for tracking and monitoring memory and resource usage."""
# Python standard Libraries
# Python standard Libraries
import logging
import socket
import os
import resource
import threading
from ast import literal_eval
from sys import argv
from datetime import datetime, timezone
import platform

# Installed libraires
from collections import defaultdict
import pandas as pd

# Geoips libraries
from geoips.filenames.base_paths import PATHS as gpaths

from geoips.utils.context_managers import import_optional_dependencies

LOG = logging.getLogger(__name__)

with import_optional_dependencies(loglevel="info"):
    """Attempt to import a package and print to LOG.info if the import fails."""
    import psutil


def print_mem_usage(logstr="", verbose=False):
    """Print memory usage to LOG.info.

    * By default include psutil output.
    * If verbose is True, include output from both psutil and resource packages.
    """
    # If psutil / socket / resource are not imported, do not fail
    usage_dict = {}
    try:
        vmem_percent = psutil.virtual_memory().percent
        LOG.info(
            "virtual perc: %s on %s %s",
            str(vmem_percent),
            str(socket.gethostname()),
            logstr,
        )
        swap_percent = psutil.swap_memory().percent
        LOG.info(
            "swap perc:    %s on %s %s",
            str(swap_percent),
            str(socket.gethostname()),
            logstr,
        )
    except NameError as resp:
        vmem_percent = swap_percent = "nan"
        LOG.info(
            "%s: psutil or socket not defined, no percent memusg output", str(resp)
        )
    try:
        highest = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        LOG.info(
            "highest:      %s on %s %s",
            str(highest),
            str(socket.gethostname()),
            logstr,
        )
    except NameError as resp:
        highest = "nan"
        LOG.info(
            "%s: resource or socket not defined, no highest memusg output", str(resp)
        )

    if verbose:
        usage_dict = print_resource_usage(logstr)
    usage_dict["memusg_virtual"] = vmem_percent
    usage_dict["memusg_swap"] = swap_percent
    usage_dict["memusg_highest"] = highest
    return usage_dict


def print_resource_usage(logstr=""):
    """Print verbose resource usage, using "resource" package."""
    usage_dict = {}
    try:
        usage = resource.getrusage(resource.RUSAGE_SELF)
        for name, desc in [
            ("ru_utime", "RESOURCE " + logstr + " User time"),
            ("ru_stime", "RESOURCE " + logstr + " System time"),
            ("ru_maxrss", "RESOURCE " + logstr + " Max. Resident Set Size"),
            ("ru_ixrss", "RESOURCE " + logstr + " Shared Memory Size"),
            ("ru_idrss", "RESOURCE " + logstr + " Unshared Memory Size"),
            ("ru_isrss", "RESOURCE " + logstr + " Stack Size"),
            ("ru_inblock", "RESOURCE " + logstr + " Block inputs"),
            ("ru_oublock", "RESOURCE " + logstr + " Block outputs"),
        ]:
            LOG.info("%-25s (%-10s) = %s", desc, name, getattr(usage, name))
            usage_dict[name] = getattr(usage, name)
    except NameError:
        LOG.info("resource not defined")
    return usage_dict


class PidLog:
    """Track a PID and all children.

    * Requires psutil and threading
    """

    def __init__(self, inpid, logstr=""):

        self.ps_parent = psutil.Process(pid=inpid)
        self.own_pid = os.getpid()
        self.logstr = logstr
        self.pid_bool = self.ps_parent.is_running()
        # allows for other process to run
        self.track_pid_thread = threading.Thread(
            target=self.track_pids, name="TrackPID", daemon=True
        )
        self.track_pid_thread.start()

    def track_pids(self):
        """Track pids and create a dict of values."""
        self.usage_dict = {
            "overall": {
                "cpu_percent": [],
                "thread_count": [],
                "unique_set_size": [],
                "res_set_size": [],
                "utc_datetime": [],
            },
            "checkpoints": {},
        }

        if platform.system() == "Linux":
            self.usage_dict["overall"]["cpu_count"] = []

        while self.pid_bool:

            ps_child = self.ps_parent.children(recursive=True)

            # filter out own pid and any children spawned
            try:
                all_pids = [self.ps_parent] + [
                    i
                    for i in ps_child
                    for j in i.parents()
                    if i.pid != self.own_pid and j.pid != self.own_pid
                ]
            except psutil.NoSuchProcess as resp:
                LOG.warning(f"Process no longer exists, skipping, {resp}")
                continue

            tmp_cpu, cpu_per, thrd_cnt, uss_tmp, rss_tmp = [], 0, 0, 0, 0

            for i in all_pids:
                try:
                    # CPU usage with oneshot doesn't work

                    with i.oneshot():
                        # cpu_num only works for Linux, FreeBSD, SunOS
                        if platform.system() == "Linux":
                            tmp_cpu += [i.cpu_num()]
                        thrd_cnt += i.num_threads()
                        tmp_mem = i.memory_full_info()
                        uss_tmp += tmp_mem.uss
                        rss_tmp += tmp_mem.rss
                    # might want average over all cpus?
                    cpu_per += i.cpu_percent(interval=0.1)
                except psutil.Error:
                    # pid expired quickly, need to remove invalid values
                    continue

            dtime = datetime.utcnow().strftime("%Y%m%d_%H%M%S_%f")
            self.usage_dict["overall"]["cpu_percent"].append(cpu_per)
            if platform.system() == "Linux":
                self.usage_dict["overall"]["cpu_count"].append(len(set(tmp_cpu)))
            self.usage_dict["overall"]["thread_count"].append(thrd_cnt)
            self.usage_dict["overall"]["unique_set_size"].append(uss_tmp)
            self.usage_dict["overall"]["res_set_size"].append(rss_tmp)
            self.usage_dict["overall"]["utc_datetime"].append(dtime)

        return

    def print_mem_usg(self, logstr="", verbose=False):
        """Print verbose resouce usage."""
        usage_dict = {}
        vmem_percent = psutil.virtual_memory().percent
        LOG.info(
            "virtual perc: %s on %s %s",
            str(vmem_percent),
            str(socket.gethostname()),
            self.logstr,
        )
        #

        swap_percent = psutil.swap_memory().percent
        LOG.info(
            "swap perc:    %s on %s %s",
            str(swap_percent),
            str(socket.gethostname()),
            self.logstr,
        )

        if len(self.usage_dict["overall"]["unique_set_size"]) > 0:
            ram_usg = max(self.usage_dict["overall"]["unique_set_size"])
            highest = max(self.usage_dict["overall"]["res_set_size"])
        else:
            ram_usg = 0
            highest = 0

        LOG.info(
            "highest ram:    %s on %s %s",
            str(ram_usg),
            str(socket.gethostname()),
            self.logstr,
        )

        LOG.info(
            "highest rss:      %s on %s %s",
            str(highest),
            str(socket.gethostname()),
            self.logstr,
        )

        if verbose:
            usage_dict = self.print_resource_usage()

        usage_dict["memusg_virtual"] = vmem_percent
        usage_dict["memusg_swap"] = swap_percent
        usage_dict["memusg_highest"] = highest

        return usage_dict

    def track_resource_usage(
        self,
        logstr="",
        verbose=False,
        key=None,
        show_log=True,
        checkpoint=False,
        increment_key=True,
    ):
        """Record resouce usage for a given processing marker in GeoIPS.

        Parameters
        ----------
        logstr : str, optional
            String to include at the start of the log message, by default ""
        verbose : bool, optional
            Print the full resource usage statistics dict to stdout, by default False
        key : str, optional
            Unique name of marker to record statistics, by default None
            A marker can be categorized as part of a group when using ":" in the
            marker name (e.g "FOO: BAR"). This is used in the checkpoint_usage_stats
            method, where markers will have a "checkpoint_group" key in their resource
            usage statistics dictionary holding the name of the group.
            (e.g. stats["BAR"]["checkpoint_group"] == "FOO")
        show_log : bool, optional
            Use LOG.info to print the current memory usage, by default True
        checkpoint : bool, optional
            Re-use a marker key to record more detailed profiling of resource usage in
            terms of both timing and maximum memory usage, by default False
        increment_key : bool, optional
            Append the number of times marker is re-used to the key, by default True
            Marker names should be unique, and should at minimum hold the start/end
            resource usage for the marker. If a marker is re-used and increment_key is
            False, a KeyError is raised. If increment_key is True and the marker has
            both start/end stats, a number is automatically appended to the key name.
            This number represents how many times a marker has been re-used. For
            example: "FOO: BAR" -> "FOO: BAR(1)": -> "FOO: BAR(2)"
            (Note: "FOO: BAR(2)" is only created if "FOO: BAR(1)" has start/end stats)

        Returns
        -------
        dict
            Resource usage statistics for checkpoint

        Raises
        ------
        KeyError
            Duplicate key is discovered and increment_key is False
        """
        usage_dict = {}
        vmem_percent = psutil.virtual_memory().percent
        swap_percent = psutil.swap_memory().percent

        overall_usage_dict = self.usage_dict["overall"]
        if len(overall_usage_dict["unique_set_size"]) > 0:
            # Take the latest value under the tracked resource dict - this might be
            # a problematic approach...
            last_stat = {x: y[-1] for x, y in overall_usage_dict.items()}
            dt = datetime.strptime(last_stat["utc_datetime"], "%Y%m%d_%H%M%S_%f")
        else:
            last_stat = {x: 0 for x in overall_usage_dict.keys()}
            dt = datetime.now(timezone.utc)
            last_stat["utc_datetime"] = dt.strftime("%Y%m%d_%H%M%S_%f")

        gran_usg_dict = self.usage_dict["checkpoints"]
        if key not in gran_usg_dict:
            LOG.info(f"New tracker {key}")
            gran_usg_dict[key] = defaultdict(list)
            elapsed_time = 0
            total_runtime = 0
        elif len(overall_usage_dict["utc_datetime"]) == 0:
            LOG.warning(
                f"UNEXPECTED ERROR IN RESOURCE TRACKING {key}: "
                "No overall usage defined, starting over with tracking :shrug:\n"
                f"gran_usg_dict[{key}] = {gran_usg_dict.get(key)}\n"
                f"overall_usage_dict: \n{overall_usage_dict}"
            )
            gran_usg_dict[key] = defaultdict(list)
            elapsed_time = 0
            total_runtime = 0
        else:
            start_dt = datetime.strptime(
                overall_usage_dict["utc_datetime"][0], "%Y%m%d_%H%M%S_%f"
            )
            prior_dt = datetime.strptime(
                gran_usg_dict[key]["utc_datetime"][-1], "%Y%m%d_%H%M%S_%f"
            )
            total_runtime = (dt - start_dt).total_seconds()
            elapsed_time = (dt - prior_dt).total_seconds()

        # Check if this key is being re-used. Arrays should at maximum length of 2:
        if (
            len(gran_usg_dict[key]["unique_set_size"]) == 2
            and checkpoint is False
            and increment_key is False
        ):
            raise KeyError(f"Usage key '{key}' already exists and is of length 2")
        elif len(gran_usg_dict[key]["unique_set_size"]) == 2 and increment_key:
            orig_key = key
            inc_keys = [x for x in gran_usg_dict.keys() if key in x]
            for inc_num, ikey in enumerate(inc_keys, 1):
                if gran_usg_dict[ikey]["unique_set_size"] == 2:
                    continue
                elif len(gran_usg_dict[ikey]["unique_set_size"]) < 2:
                    key = ikey
                    break
            else:
                key = f"{key}({inc_num})"
            if key not in gran_usg_dict:
                gran_usg_dict[key] = defaultdict(list)
                elapsed_time = 0
                total_runtime = 0
            LOG.warning(
                "Usage key '%s' already exists and is of length 2. increment_keys "
                "is True, so will instead use '%s'.",
                orig_key,
                key,
            )

        if checkpoint and len(gran_usg_dict[key]["unique_set_size"]) == 0:
            LOG.warning(
                "Flagged %s as checkpoint, but found no stats. Likely this is not "
                "a true checkpoint. Considering this the start of resource tracking.",
                key,
            )
            checkpoint = False

        if checkpoint:
            # Use sub-checkpoints to track highest memory usage
            if "subcheck" not in gran_usg_dict[key]:
                gran_usg_dict[key]["subcheck"] = defaultdict(list)
            for stat, val in last_stat.items():
                gran_usg_dict[key]["subcheck"][stat].append(val)
            return

        # Track the resource statistics for this checkpoint
        for stat, val in last_stat.items():
            gran_usg_dict[key][stat].append(val)
        # Also track the elapsed time of the checkpoint, and the total run time
        gran_usg_dict[key]["elapsed_time"].append(elapsed_time)
        gran_usg_dict[key]["total_runtime"].append(total_runtime)
        gran_usg_dict[key]["datetime"].append(dt)
        # Grab the highest value for all statistics
        if "subcheck" in gran_usg_dict[key]:
            all_stats = {
                stat: gran_usg_dict[key][stat] + gran_usg_dict[key]["subcheck"][stat]
                for stat in overall_usage_dict.keys()
            }
        else:
            all_stats = gran_usg_dict[key]
        max_stat = {
            stat: max(vals)
            for stat, vals in all_stats.items()
            if "datetime" not in stat and "max_" not in stat
        }

        for stat, val in max_stat.items():
            gran_usg_dict[key][f"max_{stat}"] = val

        ram_usg = max_stat["unique_set_size"]
        highest = max_stat["res_set_size"]
        if show_log and not checkpoint:
            LOG.info(
                "%s: virtual perc: %s on %s %s",
                key,
                str(vmem_percent),
                str(socket.gethostname()),
                self.logstr,
            )
            LOG.info(
                "%s: swap perc:    %s on %s %s",
                key,
                str(swap_percent),
                str(socket.gethostname()),
                self.logstr,
            )
            LOG.info(
                "%s: highest ram:    %s on %s %s",
                key,
                str(ram_usg),
                str(socket.gethostname()),
                self.logstr,
            )
            LOG.info(
                "%s: highest rss:      %s on %s %s",
                key,
                str(highest),
                str(socket.gethostname()),
                self.logstr,
            )
            LOG.info(
                "%s: elapsed runtime (seconds):    %s on %s %s",
                key,
                str(elapsed_time),
                str(socket.gethostname()),
                self.logstr,
            )

            if verbose:
                usage_dict = self.print_resource_usage()

        usage_dict["memusg_virtual"] = vmem_percent
        usage_dict["memusg_swap"] = swap_percent
        usage_dict["memusg_highest"] = highest
        usage_dict["elapsed_runtime"] = elapsed_time
        usage_dict["total_procflow_runtime"] = total_runtime

        return usage_dict

    def checkpoint_usage_stats(self):
        """Return organized dictionary of stats from track_resource_usage.

        Returns
        -------
        dict
            Resource usage statistics for markers/checkpoints recorded by the
            track_resource_usage method. Dictionary is ordered by the markers passed to
            track_resource_usage. Each key in the return dictionary will hold a
            dictionary of the resource usage statistics. If a group is specified in the
            marker name (e.g. "FOO: BAR"), the return dictionary will have a "BAR" key,
            and will have a "checkpoint_group" key in the corresponding resource usage
            dictionary with a value of "FOO".
        """
        checkpoint_usage_stats = {}
        exclude_stats = [
            "max_elapsed_time",
            "max_total_runtime",
            "utc_datetime",
            "subcheck",
        ]
        for check_name, check_stats in self.usage_dict["checkpoints"].items():
            checkpoint_usage_stats[check_name] = {}
            if ": " in check_name:
                group = check_name.split(": ")[0]
                checkpoint_usage_stats[check_name]["checkpoint_group"] = group
            for stat, val in check_stats.items():
                if stat in exclude_stats:
                    continue
                elif stat in ["elapsed_time"]:
                    checkpoint_usage_stats[check_name][stat] = val[-1]
                elif "max_" in stat:
                    checkpoint_usage_stats[check_name][stat] = val
                else:
                    if stat == "datetime":
                        stat = f"checkpoint_{stat}"
                    checkpoint_usage_stats[check_name][f"{stat}_start"] = val[0]
                    checkpoint_usage_stats[check_name][f"{stat}_stop"] = val[-1]
        return checkpoint_usage_stats

    def print_resource_usage(self):
        """Print verbose resource usage, using "resource" package."""
        usage_dict = {}
        try:
            usage = resource.getrusage(resource.RUSAGE_SELF)
            for name, desc in [
                ("ru_utime", "RESOURCE " + self.logstr + " User time"),
                ("ru_stime", "RESOURCE " + self.logstr + " System time"),
                ("ru_maxrss", "RESOURCE " + self.logstr + " Max. Resident Set Size"),
                ("ru_ixrss", "RESOURCE " + self.logstr + " Shared Memory Size"),
                ("ru_idrss", "RESOURCE " + self.logstr + " Unshared Memory Size"),
                ("ru_isrss", "RESOURCE " + self.logstr + " Stack Size"),
                ("ru_inblock", "RESOURCE " + self.logstr + " Block inputs"),
                ("ru_oublock", "RESOURCE " + self.logstr + " Block outputs"),
            ]:
                LOG.info("%-25s (%-10s) = %s", desc, name, getattr(usage, name))
                usage_dict[name] = getattr(usage, name)
        except NameError:
            LOG.info("resource not defined")
        usage_dict["ru_cpuusg"] = max(self.usage_dict["overall"]["cpu_percent"])  # [-1]
        usage_dict["ru_threads"] = max(self.usage_dict["overall"]["thread_count"])
        if platform.system() == "Linux":
            usage_dict["ru_cpucnt"] = max(self.usage_dict["overall"]["cpu_count"])
        usage_dict["ru_uss"] = self.usage_dict["overall"]["unique_set_size"][-1]
        usage_dict["ru_maxuss"] = max(self.usage_dict["overall"]["unique_set_size"])
        LOG.info(
            "RESOURCE "
            + self.logstr
            + " Max. User Set Size (RAM) = "
            + str(max(self.usage_dict["overall"]["unique_set_size"]))
        )
        LOG.info(
            "RESOURCE "
            + self.logstr
            + " Max. CPU % Usage = "
            + str(max(self.usage_dict["overall"]["cpu_percent"]))
        )
        return usage_dict

    def save_exit(self):
        """Exit the thread cleanly."""
        LOG.info("Stopping resource tracking.")
        self.pid_bool = False
        self.track_pid_thread.join()

    def save_csv(self):
        """Save a csv file to output."""
        usg_dict = {
            "Time [UTC]": self.usage_dict["overall"]["utc_datetime"],
            "CPU Percent": self.usage_dict["overall"]["cpu_percent"],
            "Thread Count": self.usage_dict["overall"]["thread_count"],
            "USS [RAM bytes]": self.usage_dict["overall"]["unique_set_size"],
            "RSS [bytes]": self.usage_dict["overall"]["res_set_size"],
        }
        if platform.system() == "Linux":
            usg_dict["CPU Count"] = self.usage_dict["overall"]["cpu_count"]
        df = pd.DataFrame(usg_dict)

        outdir = os.path.join(gpaths["GEOIPS_OUTDIRS"], "memory_logs")
        if not os.path.exists(outdir):
            os.mkdir(outdir)

        outname = "geoips_memtrack_p{}.csv".format(self.ps_parent.pid)
        outpath = outdir + "/" + outname
        LOG.info("Saving csv memory track to {}".format(outpath))

        df.to_csv(outpath, index=False)


def single_track_pid(procpid):
    """Output a snapshot of a pid usage on server.

    * Requires an input pid.
    """
    ps_parent = psutil.Process(pid=procpid)
    own_pid = os.getpid()
    ps_child = ps_parent.children(recursive=True)

    # filter out own pid and any children spawned
    all_pids = [ps_parent] + [
        i
        for i in ps_child
        for j in i.parents()
        if i.pid != own_pid and j.pid != own_pid
    ]

    tmp_cpu, cpu_per, thrd_cnt, uss_tmp, rss_tmp = [], 0, 0, 0, 0

    for i in all_pids:
        try:
            # CPU usage with oneshot doesn't work

            with i.oneshot():
                # cpu_num only works for Linux, FreeBSD, SunOS
                if platform.system() == "Linux":
                    tmp_cpu += [i.cpu_num()]
                thrd_cnt += i.num_threads()
                tmp_mem = i.memory_full_info()
                uss_tmp += tmp_mem.uss
                rss_tmp += tmp_mem.rss
            # might want average over all cpus?
            cpu_per += i.cpu_percent(interval=0.1)
        except psutil.Error:
            # pid expired quickly, need to remove invalid values
            continue

    if platform.system() == "Linux":
        # outputs CPU percent, CPU count, Thread count, USS, RSS
        print(
            "{}, {}, {}, {}, {}".format(
                cpu_per, len(set(tmp_cpu)), thrd_cnt, uss_tmp, rss_tmp
            )
        )
    else:
        # outputs CPU percent, Thread count, USS, RSS
        print("{}, {}, {}, {}".format(cpu_per, thrd_cnt, uss_tmp, rss_tmp))


if __name__ == "__main__":
    """Track a PID given from command line."""
    pid_track = single_track_pid(literal_eval(argv[1]))



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Data manipulation steps for "Night_Vis_IR" product.

This algorithm expects two VIIRS channels (DNBRad and M16BT) for a RGB image
"""
import logging

from geoips.image_utils.mpl_utils import alpha_from_masked_arrays, rgba_from_arrays
from geoips.data_manipulations.corrections import apply_solar_zenith_correction
from geoips.data_manipulations.corrections import apply_data_range

LOG = logging.getLogger(__name__)

interface = "algorithms"
family = "list_numpy_to_numpy"
name = "RGB_Default"


def call(
    arrays,
    sun_zen_correction=False,
    output_data_range=None,
    min_outbounds="crop",
    max_outbounds="crop",
    norm=True,
    inverse=False,
):
    """Apply default RGB algorithm.

    Plots RGB with the red gun being the first variable specified in the
    product YAML, green second, and blue third.

    Full support for this algorithm is still a WIP.

    Parameters
    ----------
    arrays : list of numpy.ndarray
        * list of numpy.ndarray or numpy.MaskedArray of channel data,
            in order of sensor "channels" list

    Returns
    -------
    numpy.ndarray
        numpy.ndarray or numpy.MaskedArray of qualitative RGBA image output
    """
    red = arrays[0]  # Red gun: First listed variable
    grn = arrays[1]  # Green gun: Second listed variable
    blu = arrays[2]  # Blue gun: Third listed variable

    if sun_zen_correction and len(arrays) >= 4:
        sun_zenith = arrays[3]

        red = apply_solar_zenith_correction(red, sun_zenith)
        grn = apply_solar_zenith_correction(grn, sun_zenith)
        blu = apply_solar_zenith_correction(blu, sun_zenith)

    if output_data_range is None:
        output_data_range = [red.min(), red.max()]
    red = apply_data_range(
        red,
        min_val=output_data_range[0],
        max_val=output_data_range[1],
        min_outbounds=min_outbounds,
        max_outbounds=max_outbounds,
        norm=norm,
        inverse=inverse,
    )
    if output_data_range is None:
        output_data_range = [grn.min(), grn.max()]
    grn = apply_data_range(
        grn,
        min_val=output_data_range[0],
        max_val=output_data_range[1],
        min_outbounds=min_outbounds,
        max_outbounds=max_outbounds,
        norm=norm,
        inverse=inverse,
    )

    if output_data_range is None:
        output_data_range = [blu.min(), blu.max()]
    blu = apply_data_range(
        blu,
        min_val=output_data_range[0],
        max_val=output_data_range[1],
        min_outbounds=min_outbounds,
        max_outbounds=max_outbounds,
        norm=norm,
        inverse=inverse,
    )

    alp = alpha_from_masked_arrays([red, grn, blu])
    rgba = rgba_from_arrays(red, grn, blu, alp)

    return rgba



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Data manipulation steps for "Volcanic-Ash" EUMETSAT RGB product.

This algorithm expects three Infrared/Visible channels for an RGB image:
* Red SEVIRI B10BT - B09BT
* Green SEVIRI B09BT - B07BT
* Blue SEVIRI B09BT
"""
import logging

LOG = logging.getLogger(__name__)

interface = "algorithms"
family = "xarray_to_numpy"
name = "Volcanic_Ash"


def call(xobj):
    """Night_Microphys product algorithm data manipulation steps.

    This algorithm expects TBs from three SEVIRI channels:

    * Red: B10BT - B09BT
    * Green: B09BT - B07BT
    * Blue: B07BT

    Parameters
    ----------
    arrays : list of numpy.ndarray
        * list of numpy.ndarray or numpy.MaskedArray of channel data,
            in order of sensor "channels" list
        * Unit in Degrees Kelvin and reflectance

    Returns
    -------
    numpy.ndarray
        numpy.ndarray or numpy.MaskedArray of qualitative RGBA image output
    """
    bt07 = xobj["B07BT"].to_masked_array()  # 3.9 um TBs, B04 SEVIRI
    bt09 = xobj["B09BT"].to_masked_array()  # 10.8 um reflectance, B09 SEVIRI
    bt10 = xobj["B10BT"].to_masked_array()  # 12.0 um reflectance, B10 SEVIRI

    # Convert TB from Kevin to Celsius
    from geoips.data_manipulations.conversions import unit_conversion

    bt07 = unit_conversion(bt07, input_units="Kelvin", output_units="celsius")
    bt09 = unit_conversion(bt09, input_units="Kelvin", output_units="celsius")
    bt10 = unit_conversion(bt10, input_units="Kelvin", output_units="celsius")

    from geoips.data_manipulations.corrections import apply_data_range, apply_gamma

    data_range = [-4, 2]
    red = apply_data_range(
        bt10 - bt09,
        min_val=data_range[0],
        max_val=data_range[1],
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=False,
    )  # need inverse option?
    red = apply_gamma(red, 1.0)
    red = apply_gamma(red, 1.0)
    # Note: may need additional solar correction

    data_range = [-4, 5]
    grn = apply_data_range(
        bt09 - bt07,
        min_val=data_range[0],
        max_val=data_range[1],
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=False,
    )
    grn = apply_gamma(grn, 1.0)
    grn = apply_gamma(grn, 1.0)
    # Note: may need additional solar correction

    data_range = [-30, 30]
    blu = apply_data_range(
        bt09,
        min_val=data_range[0],
        max_val=data_range[1],
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=False,
    )  # create image of deep clouds in blueish color
    # norm=True, inverse=False)    #create image of deep clouds in yellowish color
    blu = apply_gamma(blu, 1.0)
    blu = apply_gamma(blu, 1.0)

    from geoips.image_utils.mpl_utils import alpha_from_masked_arrays, rgba_from_arrays

    alp = alpha_from_masked_arrays([red, grn, blu])
    rgba = rgba_from_arrays(red, grn, blu, alp)

    return rgba



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Standard GeoIPS xarray dictionary based GeoKOMPSAT AMI NetCDF data reader."""

# Python Standard Libraries
from datetime import datetime, timedelta
import glob
import logging
import os
from pathlib import Path

# Third-Party Libraries
import netCDF4 as ncdf
import numpy as np
import xarray

from geoips.interfaces import readers
from geoips.filenames.base_paths import PATHS as gpaths
from geoips.utils.context_managers import import_optional_dependencies
from geoips.plugins.modules.readers.utils.geostationary_geolocation import (
    check_geolocation_cache_backend,
    get_geolocation_cache_filename,
    get_geolocation,
)

LOG = logging.getLogger(__name__)

with import_optional_dependencies(loglevel="info"):
    """Attempt to import a package & print to LOG.info if the import fails."""
    # If this reader is not installed on the system, don't fail alltogether,
    # just skip this import. This reader will not work if the import fails
    # and the package will have to be installed to process data of this type.
    import numexpr as ne
    import zarr

nprocs = 6

try:
    ne.set_num_threads(nprocs)
except Exception:
    LOG.info(
        f"Failed numexpr.set_num_threads in {__file__}. "
        "If numexpr is not installed and you need it, install it."
    )


interface = "readers"
family = "standard"
name = "ami_netcdf"
source_names = ["ami"]

# These should be added to the data file object
BADVALS = {
    "Off_Of_Disk": -999.9,
    "Conditional": -999.8,
    "Out_Of_Valid_Range": -999.7,
    "No_Value": -999.6,
    "Unitialized": -9999.9,
}

ALL_GEO_VARS = [
    "solar_zenith_angle",
    "satellite_zenith_angle",
    "solar_azimuth_angle",
    "satellite_azimuth_angle",
    "latitude",
    "longitude",
]

DATASET_INFO = {
    "MED": ["VI004", "VI005", "VI008"],
    "HIGH": ["VI006"],
    "LOW": [
        "NR013",
        "NR016",
        "SW038",
        "WV063",
        "WV069",
        "WV073",
        "IR087",
        "IR096",
        "IR105",
        "IR112",
        "IR123",
        "IR133",
    ],
}

CENTER_WAVENUMBERS = {
    "VI004": 0.4702,
    "VI005": 0.5086,
    "VI006": 0.6394,
    "VI008": 0.8630,
    "NR013": 1.3740,
    "NR016": 1.6092,
    "SW038": 2612.677373521110,
    "WV063": 1617.609242531340,
    "WV069": 1441.575428760170,
    "WV073": 1365.249992024440,
    "IR087": 1164.949392856340,
    "IR096": 1039.960216776110,
    "IR105": 966.153383926055,
    "IR112": 891.713057301260,
    "IR123": 810.609007871230,
    "IR133": 753.590621482278,
}


DONT_AUTOGEN_GEOLOCATION = False
if os.getenv("DONT_AUTOGEN_GEOLOCATION"):
    DONT_AUTOGEN_GEOLOCATION = True


"""
Equations and code from GEO-KOMPSAT-2A Level 1B Data User Manual.
"""


def latlon_from_lincol_geos(
    resolution,
    line,
    column,
    metadata,
    geolocation_cache_backend="memmap",
    chunk_size=None,
    resource_tracker=None,
):
    """Calculate latitude and longitude from array indices.

    Uses geostationary projection (likely won't work with extended local area files).
    Equations and code from GEO-KOMPSAT-2A Level 1B Data User Manual.
    """
    check_geolocation_cache_backend(geolocation_cache_backend)
    fname = get_geolocation_cache_filename(
        "GEOLL",
        metadata,
        geolocation_cache_backend=geolocation_cache_backend,
        chunk_size=chunk_size,
    )
    if resource_tracker is not None:
        key = Path(fname).name
        if metadata.get("area_id"):
            key += f"_{metadata.get('area_id')}"
        resource_tracker.track_resource_usage(logstr="MEMUSG", verbose=False, key=key)
    if not os.path.isfile(fname):
        degtorad = 3.14159265358979 / 180.0

        # Note metadata included directly in the data files is incorrect for
        # some AMI data files.  These hard coded values are pulled from the
        # manual for correct geolocation.
        if resolution == "HIGH":
            COFF = 11000.5
            CFAC = 8.170135561335742e7
            LOFF = 11000.5
            LFAC = 8.170135561335742e7
        elif resolution == "MED":
            COFF = 5500.5
            CFAC = 4.0850677806678705e7
            LOFF = 5500.5
            LFAC = 4.0850677806678705e7
        else:
            COFF = 2750.5
            CFAC = 2.0425338903339352e7
            LOFF = 2750.5
            LFAC = 2.0425338903339352e7
        sub_lon = 128.2

        # COFF = metadata["COFF"]
        # CFAC = metadata["CFAC"]
        # LOFF = metadata["LOFF"]
        # LFAC = metadata["LFAC"]
        # sub_lon = metadata["sub_lon"]

        sub_lon = sub_lon * degtorad

        x = np.empty_like(column)
        y = np.empty_like(line)
        cosx = np.empty_like(x)
        cosy = np.empty_like(x)
        sinx = np.empty_like(x)
        siny = np.empty_like(x)
        cosxy = np.empty_like(x)
        A = np.empty_like(x)
        Sd = np.empty_like(x)
        Sn = np.empty_like(x)
        S1 = np.empty_like(x)
        S2 = np.empty_like(x)
        S3 = np.empty_like(x)
        Sxy = np.empty_like(x)

        x = degtorad * ((column - COFF) * 2**16 / CFAC)
        y = degtorad * ((line - LOFF) * 2**16 / LFAC)
        x = x.astype(np.float32)
        y = y.astype(np.float32)
        ne.evaluate("cos(x)", out=cosx)
        ne.evaluate("cos(y)", out=cosy)
        ne.evaluate("sin(x)", out=sinx)
        ne.evaluate("sin(y)", out=siny)
        ne.evaluate("cosx * cosy", out=cosxy)
        ne.evaluate("42164 * cosxy", out=A)
        B = cosy**2 + 1.006739501 * siny**2
        B = B.astype(np.float32)
        ne.evaluate("sqrt(A**2 - B*1737122264)", out=Sd)
        ne.evaluate("(A - Sd) / B", out=Sn)
        ne.evaluate("42164 - (Sn * cosxy)", out=S1)
        ne.evaluate("Sn * (sinx * cosy)", out=S2)
        ne.evaluate("-Sn * siny", out=S3)
        ne.evaluate("sqrt((S1*S1) + (S2*S2))", out=Sxy)
        nlon = (np.arctan(S2 / S1) + sub_lon) / degtorad
        nlat = np.arctan((1.006739501 * S3) / Sxy) / degtorad

        nlat[np.where(np.isnan(nlat))] = -999.9
        nlon[np.where(np.isnan(nlon))] = -999.9
        if geolocation_cache_backend == "memmap":
            with open(fname, "w") as df:
                nlat.tofile(df)
                nlon.tofile(df)
        elif geolocation_cache_backend == "zarr":
            if chunk_size:
                chunks = (chunk_size, chunk_size)
            else:
                chunks = None
            LOG.info("Storing to %s (chunks=%s)", fname, chunks)
            # zarr does NOT have a close method, so you can NOT use the with context.
            zf = zarr.open(fname, mode="w")
            # Assume both arrays have the same shape and dtype
            kwargs = {
                "shape": nlat.shape,
                "dtype": nlat.dtype,
            }
            # As of Python 3.11, can't pass chunks=None into create_dataset
            if chunks:
                kwargs["chunks"] = chunks
            zf.create_dataset("lats", **kwargs)
            zf.create_dataset("lons", **kwargs)
            zf["lats"][:] = nlat
            zf["lons"][:] = nlon
    else:
        # Create memmap to the lat/lon file
        # Nothing will be read until explicitly requested
        # We are mapping this here so that the lats and lons are available when
        # calculating satlelite angles
        if geolocation_cache_backend == "memmap":
            shape = (metadata["num_lines"], metadata["num_samples"])
            offset = 4 * metadata["num_samples"] * metadata["num_lines"]
            nlat = np.memmap(fname, mode="r", dtype=np.float32, offset=0, shape=shape)
            nlon = np.memmap(
                fname, mode="r", dtype=np.float32, offset=offset, shape=shape
            )
        elif geolocation_cache_backend == "zarr":
            LOG.info(
                "GETGEO zarr to {} : lat/lon file for {}".format(
                    fname, metadata["scene"]
                )
            )
            zf = zarr.open(fname, mode="r")
            nlat = zf["lats"]
            nlon = zf["lons"]
    if resource_tracker is not None:
        resource_tracker.track_resource_usage(logstr="MEMUSG", verbose=False, key=key)

    return (nlat, nlon)


# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
# Metadata collection functions.
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
def _check_file_consistency(metadata):
    """
    Check that all input metadata are from the same image time.

    Performs cheks on satellite_name, instrument_name, observation_mode, and time.
    If these are all equal, returns True.
    If any differ, returns False.
    """
    # Checking file-level metadata for exact equality in the following fields
    checks = {
        "satellite_name": "general",
        "instrument_name": "general",
        "observation_mode": "observation",
        "observation_start_time": "observation",
    }
    for name, level in checks.items():
        check_set = set(
            [metadata[fname]["global"][level][name] for fname in metadata.keys()]
        )
        if len(check_set) != 1:
            LOG.debug("Failed on {0}. Found: {1}".format(name, check_set))
            return False
    return True


def _get_general_metadata(df):
    """Gather all of the general file-level metadata."""
    metadata = {}
    md_names = [
        "satellite_name",
        "instrument_name",
        "data_processing_center",
        "data_processing_mode",
        "channel_spatial_resolution",
        "channel_center_wavelength",
        "scene_acquisition_time",
        "mission_reference_time",
        "file_generation_time",
        "file_name",
        "file_format_version",
        "geometric_correction_sw_version",
        "star_catalog_version",
        "landmark_catalog_version",
    ]

    for name in md_names:
        if hasattr(df, name):
            metadata[name] = getattr(df, name)
        else:
            LOG.info("Warning! File-level metadata field missing: {0}".format(name))
    return metadata


def _get_data_metadata(df):
    """Gather all of the data-size metadata."""
    metadata = {}
    md_names = [
        "total_pixel_data_size",
        "number_of_total_swaths",
        "number_of_columns",
        "number_of_lines",
    ]

    for name in md_names:
        if hasattr(df, name):
            metadata[name] = getattr(df, name)
        else:
            LOG.info("Warning! Data metadata field missing: {0}".format(name))
    return metadata


def _get_observation_metadata(df):
    """Gather all of the observation metadata."""
    metadata = {}
    md_names = [
        "observation_mode",
        "observation_start_time",
        "observation_end_time",
        "time_synchro_obt",
        "time_synchro_utc",
    ]

    for name in md_names:
        if hasattr(df, name):
            metadata[name] = getattr(df, name)
        else:
            LOG.info("Warning! Observation metadata field missing: {0}".format(name))
    return metadata


def _get_projection_metadata(df):
    """Gather all of the projection metadata."""
    metadata = {}
    md_names = [
        "projection_type",
        "sub_longitude",
        "cfac",
        "lfac",
        "coff",
        "loff",
        "nominal_satellite_height",
        "earth_equatorial_radius",
        "earth_polar_radius",
        "earth_ellipsoid_center",
        "image_upperleft_latitude",
        "image_upperleft_longitude",
        "image_lowerright_latitude",
        "image_lowerright_longitude",
        "image_center_latitude",
        "image_center_longitude",
        "image_upperleft_x",
        "image_upperleft_y",
        "image_lowerright_x",
        "image_lowerright_y",
        "resampling_kernel_type",
    ]

    for name in md_names:
        if hasattr(df, name):
            metadata[name] = getattr(df, name)
        else:
            LOG.info("Warning! Projection metadata field missing: {0}".format(name))
    return metadata


def _get_global_metadata(df):
    """Get file-level metadata."""
    metadata = {}
    metadata["general"] = _get_general_metadata(df)
    metadata["data"] = _get_data_metadata(df)
    metadata["observation"] = _get_observation_metadata(df)
    metadata["projection"] = _get_projection_metadata(df)
    return metadata


def _get_variable_metadata(df):
    """Gather all of the variable metadata."""
    metadata = {}
    md_names = [
        "channel_name",
        "detector_side",
        "number_of_total_pixels",
        "number_of_error_pixels",
        "max_pixel_value",
        "max_pixel_value",
        "average_pixel_value",
        "stddev_pixel_value",
        "number_of_total_bits_per_pixel",
        "number_of_data_quality_flag_bits_per_pixel",
        "number_of_valid_bits_per_pixel",
        "data_quality_flag_meaning",
        "ground_sample_distance_ew",
        "ground_sample_distance_ns",
    ]

    for name in md_names:
        if hasattr(df.variables["image_pixel_values"], name):
            metadata[name] = getattr(df.variables["image_pixel_values"], name)
        else:
            LOG.info("Warning! Projection metadata field missing: {0}".format(name))
    return metadata


def _get_metadata(df, fname):
    """
    Gather metadata for the data file and return as a dictionary.

    Note: We are gathering all of the available metadata in case it is needed at
    some point.
    """
    metadata = {}
    # Gather all file-level metadata.
    metadata["global"] = _get_global_metadata(df)
    # Gather all variable-level metadata.
    metadata["var_info"] = _get_variable_metadata(df)
    # Gather some useful info to the top level.
    try:
        metadata["path"] = df.filepath()
    except ValueError:
        # Without cython installed, df.filepath() does not work
        metadata["path"] = fname
    metadata["satellite"] = metadata["global"]["general"]["satellite_name"]
    metadata["sensor"] = df.instrument_name
    # These were being returned as np.int32's, causing an overflow when passed
    # to numpy 2.x memmap call in latlon_from_lincol_geos function below.
    # Explicitly cast to np.int64 to avoid overflow in memmap
    metadata["num_lines"] = np.int64(df.number_of_lines)
    metadata["num_samples"] = np.int64(df.number_of_columns)
    return metadata


def metadata_to_datetime(metadata):
    """Use information from the metadata to get the image datetime."""
    start_time = metadata["global"]["observation"]["observation_start_time"]
    end_time = metadata["global"]["observation"]["observation_end_time"]
    epoch = datetime(2000, 1, 1, 12, 0, 0)
    start_time = epoch + timedelta(seconds=start_time)
    end_time = epoch + timedelta(seconds=end_time)
    return start_time, end_time


def _get_geolocation_metadata(metadata):
    """
    Gather all of the metadata used in creating geolocation data for input file.

    This is split out so we can easily create a chash of the data for creation
    of a unique filename. This allows us to avoid recalculation of angles that
    have already been calculated.
    """
    metadata = metadata["global"]
    geomet = {}
    geomet["platform_name"] = metadata["general"]["satellite_name"]
    geomet["Re"] = metadata["projection"]["earth_equatorial_radius"]
    geomet["Rp"] = metadata["projection"]["earth_polar_radius"]
    geomet["e"] = 0.0818191910435
    # Nominal satellite height is from the center of the Earth.
    geomet["H_m"] = metadata["projection"]["nominal_satellite_height"]
    geomet["pphgt"] = geomet["H_m"] - geomet["Re"]
    geomet["lat0"] = metadata["projection"]["image_center_latitude"]
    geomet["lon0"] = metadata["projection"]["image_center_longitude"] * 180.0 / np.pi
    geomet["scene"] = metadata["observation"]["observation_mode"]
    # Just getting the nadir resolution in kilometers.  Must extract from a string.
    geomet["res_km"] = float(metadata["general"]["channel_spatial_resolution"])
    geomet["roi_factor"] = 5  # roi = res * roi_factor, was 10
    # These were being returned as np.int32's, causing an overflow when passed
    # to numpy 2.x memmap call in latlon_from_lincol_geos function below.
    # Explicitly cast to np.int64 to avoid overflow in memmap
    geomet["num_lines"] = np.int64(metadata["data"]["number_of_lines"])
    geomet["num_samples"] = np.int64(metadata["data"]["number_of_columns"])
    # Dynamically get offsets and scale factors needed for correct geolocation
    # calculation
    geomet["sub_lon"] = metadata["projection"]["sub_longitude"]
    geomet["CFAC"] = np.abs(metadata["projection"]["cfac"])
    geomet["COFF"] = np.abs(metadata["projection"]["coff"])
    geomet["LFAC"] = np.abs(metadata["projection"]["lfac"])
    geomet["LOFF"] = np.abs(metadata["projection"]["loff"])
    return geomet


# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
# Get pixel data.
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
def get_data(gvars, fname, rad=False, ref=False, bt=False):
    """Get actual pixel data from a specific file using geolocation information."""
    if "Lines" in gvars and "Samples" in gvars:
        full_disk = False
        line_inds = gvars["Lines"]
        sample_inds = gvars["Samples"]
    else:
        full_disk = True

    ds = ncdf.Dataset(fname, "r")
    ipixel = ds.variables["image_pixel_values"]

    if not full_disk:
        counts = ipixel[:].data[line_inds, sample_inds]
    else:
        # Leaving out zoom stuff in abi reader for now, since it seems
        #   those conditions shouldn't be reached for these files.
        counts = ipixel[:].data

    # The following section of bitwise "masking" is from the manual.
    # I believe what is expected to happen is that, if a pixel is out-of-area
    # or erroneous, it will be exactly 32768 or 49152, and will be 0 after the
    # masking. Conditional pixels will be treated as "good enough" and will have
    # a value less than 16384 (i.e. the "normal data" range after the bitwise
    # mask is applied.
    channel = ipixel.getncattr("channel_name")
    if (channel == "VI004") or (channel == "VI005") or (channel == "NR016"):
        mask = 0b0000011111111111  # 11bit mask
    elif (channel == "VI006") or (channel == "NR013") or (channel == "WV063"):
        mask = 0b0000111111111111  # 12bit mask
    elif channel == "SW038":
        mask = 0b0011111111111111  # 14bit mask
    else:
        mask = 0b0001111111111111  # 13bit mask

    if channel.startswith("VI") or channel.startswith("NR"):
        reflectance_chan = True
    else:
        reflectance_chan = False

    counts_masked = np.bitwise_and(counts, mask)

    # So far, looks like the binary masking from the manual is sufficient.
    # (Good and conditional values will be kept.)
    # The following code is a starting point if conditional values need to be
    # masked.
    # conditional_threshold = 16384
    # out_of_area_threshold = 32768
    # erroneous_threshold = 49152
    # counts_masked = np.ma.masked_where(counts >= conditional_threshold, counts)

    gain = ds.DN_to_Radiance_Gain
    offset = ds.DN_to_Radiance_Offset

    # Equations from last page of manual (Table 14)
    # radiance = gain * count + offset
    radiance = gain * counts_masked + offset

    data = {}
    if rad:
        data["Rad"] = radiance

    if reflectance_chan and ref:
        # Visible and Near IR
        c_prime = ds.Radiance_to_Albedo_c
        albedo = radiance * c_prime
        data["Ref"] = albedo

    if bt and not reflectance_chan:
        # IR
        # Using hard-coded values for nu, since the values obtained from the
        #   dataset don't quite match the values from the manual.
        nu = CENTER_WAVENUMBERS[channel]
        h = ds.Plank_constant_h
        k = ds.Boltzmann_constant_k
        c = ds.light_speed
        c0 = ds.Teff_to_Tbb_c0
        c1 = ds.Teff_to_Tbb_c1
        c2 = ds.Teff_to_Tbb_c2
        T_e = (h * c / k * (nu * 100)) / np.log(
            (2 * h * c**2) * (nu * 100) ** 3 / (radiance * 1e-5) + 1
        )
        T_b = c0 + c1 * T_e + c2 * T_e**2

        data["BT"] = T_b

    ds.close()

    return data


def call(
    fnames,
    metadata_only=False,
    chans=None,
    area_def=None,
    self_register=False,
    roi=None,
    geolocation_cache_backend=gpaths["GEOIPS_GEOLOCATION_CACHE_BACKEND"],
    cache_chunk_size=None,
    cache_solar_angles=False,
    geolocation_only=False,
    resource_tracker=None,
    satellite_zenith_angle_cutoff=75,
):
    """
    Read Geo-Kompsat NetCDF data from a list of filenames.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.
    roi: radius of influence (unit in meter), used in interpolation scheme
        * Default=None, i.e., if not defined in the yaml file
        * Defined in yaml file where variables and tuning parameters are set
    satellite_zenith_angle_cutoff: cutoff for satellite zenith angle in degrees
        * Defaults to 75 for now to be consistent with previous hard coded
          value
        * Later will update to None (will have to update test scripts or
          outputs accordingly)

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    return readers.read_data_to_xarray_dict(
        fnames,
        call_single_time,
        metadata_only,
        chans,
        area_def,
        self_register,
        roi=roi,
        geolocation_cache_backend=geolocation_cache_backend,
        cache_chunk_size=cache_chunk_size,
        cache_solar_angles=cache_solar_angles,
        geolocation_only=geolocation_only,
        resource_tracker=resource_tracker,
        satellite_zenith_angle_cutoff=satellite_zenith_angle_cutoff,
    )


def call_single_time(
    fnames,
    metadata_only=False,
    chans=None,
    area_def=None,
    self_register=False,
    roi=None,
    geolocation_cache_backend="memmap",
    cache_chunk_size=None,
    cache_solar_angles=False,
    geolocation_only=False,
    resource_tracker=None,
    satellite_zenith_angle_cutoff=75,
):
    """
    Read Geo-Kompsat NetCDF data from a list of filenames.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.
    roi: radius of influence (unit in meter), used in interpolation scheme
        * Passed in from Call function
        * Default=None, i.e., if not defined in the yaml file
        * Defined in yaml file where variables and tuning parameters are set
    geolocation_cache_backend : str
        * Specify to use either memmap or zarray to store pre-calculated geolocation
          data.
    cache_chunk_size : int
        * Specify chunk size if using zarray to store pre-calculated geolocation data.
    resource_tracker: geoips.utils.memusg.PidLog object
        * Track resource usage using the PidLog class object from geoips.utils.memusg.
        * The PidLog.track_resource_usage method allows us to snapshot the memory usage
          for the PID associated with the geoips call. The time and stats of the
          snapshot are recorded, and can be accessed using the
          PidLog.checkpoint_usage_stats method.
    satellite_zenith_angle_cutoff: cutoff for satellite zenith angle in degrees
        * Defaults to None

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    check_geolocation_cache_backend(geolocation_cache_backend)
    gvars = {}
    datavars = {}
    geo_metadata = {}

    if fnames is None or len(fnames) == 0:
        raise ValueError("No input files specified")

    # Get metadata for all input data files
    # Check to be sure that all input files are form the same image time
    all_metadata = {}
    for fname in fnames:
        if chans:
            gotone = False
            for chan in chans:
                currchan = chan[:5].lower()
                if currchan in fname:
                    gotone = True
            if not gotone:
                LOG.info(
                    "SKIPPING file %s, not needed from channel list %s", fname, chans
                )
                continue
        try:
            # Open using with to avoid seg faults due to not properly closing files.
            # This did not seg fault prior to netcdf 1.7 / numpy 2.x
            with ncdf.Dataset(str(fname), "r") as ds:
                all_metadata[fname] = _get_metadata(ds, fname)
        except IOError as resp:
            LOG.exception("BAD FILE %s skipping", resp)
            continue
        if metadata_only:
            LOG.info("Only need metadata from first file, returning")
            break

    if not _check_file_consistency(all_metadata):
        raise ValueError("Input files inconsistent.")

    # Now put together a dict that shows what we actually got
    # This is largely to help us read in order and only create arrays of the minimum
    # required size.
    # Dict structure is channel{metadata}
    file_info = {}

    xarray_obj = xarray.Dataset()
    for fname, md in all_metadata.items():
        ch = md["var_info"]["channel_name"]
        if ch not in file_info:
            file_info[ch] = md

    xarray_obj.attrs["file_metadata"] = file_info.copy()

    # Most of the metadata are the same between files.
    # From here on we will just rely on the metadata from a single data file
    # for each resolution
    res_md = {}
    for res in ["LOW", "MED", "HIGH"]:
        # Find a file for this resolution: Any one will do
        res_chans = list(set(DATASET_INFO[res]).intersection(file_info.keys()))
        if res_chans:
            res_md[res] = file_info[res_chans[0]]

    # If we plan to self register, make sure we requested a resolution that we actually
    # plan to read.
    #
    # This could be problematic if we try to self-register to LOW when only
    # reading MED or something
    if self_register and self_register not in res_md:
        raise ValueError(
            "Resolution requested for self registration has not been read."
        )

    # Gather metadata
    # Assume the same for all resolutions.  Not true, but close enough.
    highest_md = res_md[list(res_md.keys())[0]]
    start_dt, end_dt = metadata_to_datetime(highest_md)
    xarray_obj.attrs["start_datetime"] = start_dt
    xarray_obj.attrs["end_datetime"] = end_dt
    xarray_obj.attrs["source_name"] = "ami"  # "ami"
    xarray_obj.attrs["data_provider"] = "nmsc"
    xarray_obj.attrs["platform_name"] = highest_md["global"]["general"][
        "satellite_name"
    ]
    xarray_obj.attrs["area_definition"] = area_def
    xarray_obj.attrs["source_file_names"] = [
        os.path.basename(fname) for fname in fnames
    ]

    # Get appropriate sector name
    if area_def:
        xarray_obj.attrs["area_id"] = area_def.area_id
        xarray_obj.attrs["area_definition"] = area_def
    else:
        if self_register and self_register not in DATASET_INFO:
            raise ValueError(
                f"Unrecognized resolution name requested for self "
                f"registration: {self_register}"
            )
        scene_id = highest_md["global"]["observation"]["observation_mode"].lower()
        if scene_id == "fd":
            xarray_obj.attrs["area_id"] = "Full-Disk"
        elif scene_id == "ela":
            xarray_obj.attrs["area_id"] = "Extended-Local-Area"
        elif scene_id == "la":
            xarray_obj.attrs["area_id"] = "Local-Area"
        elif scene_id == "enh":
            xarray_obj.attrs["area_id"] = "Extend-Northern-Hemisphere"
        elif scene_id == "ea":
            xarray_obj.attrs["area_id"] = "East-Asia"
        elif scene_id == "ko":
            xarray_obj.attrs["area_id"] = "Korea"
        elif scene_id == "tp":
            xarray_obj.attrs["area_id"] = "Typhoon"
        xarray_obj.attrs["area_definition"] = None
    adname = xarray_obj.attrs["area_id"]

    if metadata_only:
        LOG.info("Only need metadata from first file, returning")
        return {"METADATA": xarray_obj}

    # Create list of all possible channels for the case where no channels were requested
    all_chans_list = []
    for chan_list in DATASET_INFO.values():
        for chl in chan_list:
            all_chans_list.append(chl + "Rad")
            if chl.startswith("VI") or chl.startswith("NR"):
                all_chans_list.append(chl + "Ref")
            else:
                all_chans_list.append(chl + "BT")

    # If specific channels were requested, check them against the input data
    # If specific channels were requested, but no files exist for one of the
    # channels, then error
    if chans:
        for chan in chans:
            if chan in ALL_GEO_VARS:
                continue
            if chan not in all_chans_list:
                raise ValueError("Requested channel {0} not recognized.".format(chan))
            if chan[0:5] not in file_info.keys():
                continue

    # If no specific channels were requested, get everything
    if not chans:
        chans = all_chans_list

    # Creates a dictionary whose keys are band numbers in the form {name}##
    # and whose values are lists containing the data types(s) requested for
    # the band (e.g. Rad, Ref, BT).
    chan_info = {}
    for ch in chans:
        if ch in ALL_GEO_VARS:
            continue
        chn = ch[0:5]
        typ = ch[5:]
        if chn not in chan_info:
            chan_info[chn] = []
        chan_info[chn].append(typ)

    # Gather geolocation data
    # Assume datetime the same for all resolutions.  Not true, but close enough.
    # This saves us from having very slightly different solar angles for each channel.
    # Loop over resolutions and get metadata as needed
    if self_register:
        LOG.info("Getting geolocation information for adname %s.", adname)
        geo_metadata[adname] = _get_geolocation_metadata(res_md[self_register])

        i = np.arange(0, geo_metadata[adname]["num_lines"], dtype="f")
        j = np.arange(0, geo_metadata[adname]["num_samples"], dtype="f")
        i, j = np.meshgrid(i, j)
        (fldk_lats, fldk_lons) = latlon_from_lincol_geos(
            self_register,
            j,
            i,
            geo_metadata[adname],
            geolocation_cache_backend=geolocation_cache_backend,
            resource_tracker=resource_tracker,
        )

        gvars[adname] = get_geolocation(
            start_dt,
            geo_metadata[adname],
            fldk_lats,
            fldk_lons,
            BADVALS,
            area_def,
            geolocation_cache_backend=geolocation_cache_backend,
            cache_solar_angles=cache_solar_angles,
            resource_tracker=resource_tracker,
        )
        if not gvars[adname]:
            LOG.error(
                "GEOLOCATION FAILED for adname %s DONT_AUTOGEN_GEOLOCATION is: %s",
                adname,
                DONT_AUTOGEN_GEOLOCATION,
            )
            gvars[adname] = {}
    else:
        for res in ["LOW", "MED", "HIGH"]:
            try:
                res_md[res]
            except KeyError:
                continue
            LOG.info(
                "Getting geolocation information for resolution %s for %s", res, adname
            )
            try:
                geo_metadata[res] = _get_geolocation_metadata(res_md[res])

                i = np.arange(0, geo_metadata[res]["num_lines"], dtype="f")
                j = np.arange(0, geo_metadata[res]["num_samples"], dtype="f")
                i, j = np.meshgrid(i, j)
                (fldk_lats, fldk_lons) = latlon_from_lincol_geos(
                    res,
                    j,
                    i,
                    geo_metadata[res],
                    geolocation_cache_backend=geolocation_cache_backend,
                    resource_tracker=resource_tracker,
                )
                gvars[res] = get_geolocation(
                    start_dt,
                    geo_metadata[res],
                    fldk_lats,
                    fldk_lons,
                    BADVALS,
                    area_def,
                    geolocation_cache_backend=geolocation_cache_backend,
                    cache_solar_angles=cache_solar_angles,
                    resource_tracker=resource_tracker,
                )
            except IndexError as resp:
                LOG.exception("SKIPPING apparently no coverage or bad geolocation file")
                raise IndexError(resp)

    LOG.interactive("Done with geolocation for {}".format(adname))
    if geolocation_only:
        geo_xarrays = {"METADATA": xarray_obj}
        for dsname in gvars.keys():
            geo_xarrays[dsname] = xarray.Dataset()
            geo_xarrays[dsname].attrs = xarray_obj.attrs.copy()
            for varname in gvars[dsname].keys():
                geo_xarrays[dsname][varname] = xarray.DataArray(gvars[dsname][varname])
        return geo_xarrays

    # Read the data
    # Will read all data if sector_definition is None
    # Will only read required data if an sector_definition is provided
    for chan, types in chan_info.items():
        # If we didn't pass all the channel data, skip non-existent data types
        if chan not in file_info:
            continue
        LOG.info("Reading {}".format(chan))
        if resource_tracker:
            key = f"READ CHAN: ami_netcdf_chan_{chan}_{adname}"
            resource_tracker.track_resource_usage(
                logstr="MEMUSG",
                verbose=False,
                key=key,
                show_log=False,
            )
        chan_md = file_info[chan]
        for res, res_chans in DATASET_INFO.items():
            if chan in res_chans:
                break
        if (not self_register) and (
            res not in gvars.keys() or not gvars[res] or "latitude" not in gvars[res]
        ):
            LOG.info(
                f"We don't have geolocation information for {res} for {adname} "
                f"skipping {chan}."
            )
            continue
        if not area_def:
            dsname = res
        else:
            dsname = adname

        if self_register:
            gname = adname
        else:
            gname = res

        rad = ref = bt = False
        if "Rad" in types:
            rad = True
        if "Ref" in types:
            ref = True
        if "BT" in types:
            bt = True

        data = get_data(gvars[gname], chan_md["path"], rad, ref, bt)
        for typ, val in data.items():
            if dsname not in datavars:
                datavars[dsname] = {}
            datavars[dsname][chan + typ] = val
        if resource_tracker:
            resource_tracker.track_resource_usage(
                logstr="MEMUSG", verbose=False, key=key
            )

    if area_def:
        for res in ["LOW", "MED", "HIGH"]:
            if res in gvars and gvars[res]:
                gvars_res = gvars.pop(res)
                if adname not in gvars:
                    gvars[adname] = gvars_res

    if self_register:
        # Determine which resolution has geolocation
        LOG.info("Registering to {}".format(self_register))
        all_res = ["LOW", "MED", "HIGH"]
        if self_register not in all_res:
            raise ValueError("No geolocation data found.")

        all_res.remove(self_register)
        datavars[adname] = datavars.pop(self_register)
        for res in all_res:
            if res in datavars:
                for varname, var in datavars[res].items():
                    datavars[adname][varname] = var
                datavars.pop(res)

    # Remove lines and samples arrays.  Not needed.
    for res in gvars.keys():
        # Providing None as a default in case the dictionary is missing 'Lines' or
        # 'Samples'. Since we skip if these don't exist anyways, I don't see this
        # being an issue.

        # The reason this is useful is for cases in which these variables are
        # missing but you still want to mask all of the other variables in this
        # resolution by the satellite zenith angle cutoff provided.
        gvars[res].pop("Lines", None)
        gvars[res].pop("Samples", None)
        for varname, var in gvars[res].items():
            if satellite_zenith_angle_cutoff:
                LOG.info(
                    "Masking var %s greater than %s degrees sat zenith angle",
                    varname,
                    satellite_zenith_angle_cutoff,
                )
                gvars[res][varname] = np.ma.array(
                    var, mask=gvars[res]["satellite_zenith_angle"].mask
                )
                gvars[res][varname] = np.ma.masked_where(
                    gvars[res]["satellite_zenith_angle"]
                    > satellite_zenith_angle_cutoff,
                    gvars[res][varname],
                )
    for ds in datavars.keys():
        if not datavars[ds]:
            datavars.pop(ds)

    # Create the final dictionary of xarray objects.
    xarray_objs = {}
    for dsname in datavars.keys():
        # Create a new xarray object to populate with all the data.
        xobj = xarray.Dataset()
        xobj.attrs = xarray_obj.attrs.copy()

        for varname in datavars[dsname].keys():
            xobj[varname] = xarray.DataArray(datavars[dsname][varname])

        for varname in gvars[dsname].keys():
            xobj[varname] = xarray.DataArray(gvars[dsname][varname])

        # if roi is not defined in yaml file, a default roi is applied
        if roi is None:
            roi = 500
            if hasattr(xobj, "area_definition") and xobj.area_definition is not None:
                roi = max(
                    xobj.area_definition.pixel_size_x, xobj.area_definition.pixel_size_y
                )
                LOG.info("Trying area_def roi %s", roi)
            for curr_res in geo_metadata.keys():
                LOG.info(
                    "Trying metadata roi %s %s",
                    geo_metadata[curr_res]["res_km"] * 1000.0,
                    roi,
                )
                if geo_metadata[curr_res]["res_km"] * 1000.0 > roi:
                    roi = geo_metadata[curr_res]["res_km"] * 1000.0
                    LOG.info("Trying standard_metadata[%s] %s", curr_res, roi)
        xobj.attrs["interpolation_radius_of_influence"] = roi
        xobj.attrs["longitude_of_projection_origin"] = geo_metadata[curr_res][
            "lon0"
        ].round(5)
        xarray_objs[dsname] = xobj
        # At some point we may need to deconflict, but for now just use any of the
        # dataset attributes as the METADATA dataset
        xarray_objs["METADATA"] = xobj[[]]

    LOG.info("Done reading GEOKOMPSAT AMI data for %s", adname)

    return xarray_objs


# Unit test functions
def get_test_files(test_data_dir):
    """Generate testing xarray from test data."""
    filepath = test_data_dir + "/test_data_ami/data/20231208_0300_daytime/*.nc"
    filelist = glob.glob(filepath)
    if len(filelist) == 0:
        raise FileNotFoundError("No files found")
    for file in filelist:
        if not os.path.exists(file):
            raise FileNotFoundError(f"File {file} does not exist")
    tmp_xr = call(filelist)
    return tmp_xr


def get_test_parameters():
    """Generate test data key for unit testing."""
    return [
        {"data_key": "HIGH", "data_var": "VI006Ref", "mean": 0.14788916},
        {"data_key": "LOW", "data_var": "IR112BT", "mean": 286.63754873},
    ]



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""TC trackfile parser for F-Deck formatted TC deck files.

Each F-Deck file contains the current, official forecast for a storm. One storm
location per line (split between 2 lines in comments below for readability)::
Format is very similar to B-Deck file but line length is smaller

  AL, 99, 2023082312, 03, OFCL,   0, 187N,  651W,  25,    0, TD,  34, NEQ,
   0,    0,    0,    0,    0,    0,   0,  35,   0,    ,   0, ASL, 300,  13,
  AL, 99, 2023082312, 03, OFCL,   3, 188N,  658W,  25, 1009, TD,  34, NEQ,
   0,    0,    0,    0,    0,    0,   0,  35,   0,    ,   0, ASL, 285,  13,
  AL, 99, 2023082312, 03, OFCL,  12, 190N,  678W,  25,    0, LO,  34, NEQ,
   0,    0,    0,    0,    0,    0,   0,  35,   0,    ,   0, ASL, 275,  13,
  AL, 99, 2023082312, 03, OFCL,  24, 191N,  692W,  25,    0, LO,  34, NEQ,
   0,    0,    0,    0,    0,    0,   0,  35,   0,    ,   0, ASL, 270,  16,
  AL, 99, 2023082312, 03, OFCL,  36, 193N,  714W,  25,    0, LO,  34, NEQ,
   0,    0,    0,    0,    0,    0,   0,  35,   0,    ,   0, ASL, 270,  16,

or

  WP, 17, 2023111306, 03, JTWC,   0,  81N, 1390E,  25,    0, TD,  34, NEQ,
   0,    0,    0,    0,
  WP, 17, 2023111306, 03, JTWC,  12,  78N, 1386E,  25,    0, TD,  34, NEQ,
   0,    0,    0,    0,
  WP, 17, 2023111306, 03, JTWC,  24,  79N, 1380E,  25,    0, TD,  34, NEQ,
   0,    0,    0,    0,
  WP, 17, 2023111306, 03, JTWC,  36,  80N, 1371E,  30,    0, TD,  34, NEQ,
   0,    0,    0,    0,
  WP, 17, 2023111306, 03, JTWC,  48,  82N, 1356E,  30,    0, TD,  34, NEQ,
   0,    0,    0,    0,
  WP, 17, 2023111306, 03, JTWC,  72,  89N, 1326E,  35,    0, TS,  34, NEQ,
   60,   30,   50,   70,
  WP, 17, 2023111306, 03, JTWC,  96, 101N, 1289E,  40,    0, TS,  34, NEQ,
   70,   70,   50,   80,    0,    0,   0,  50,   0,
  WP, 17, 2023111306, 03, JTWC, 120, 113N, 1257E,  45,    0, TS,  34, NEQ,
   100,  120,   90,  120,    0,    0,   0,  55,   0,

"""
import os
import logging
from datetime import datetime, timedelta

LOG = logging.getLogger(__name__)

interface = "sector_metadata_generators"
family = "tc"
name = "fdeck_parser"


def call(trackfile_name, allowed_aid_types=None):
    """TC deckfile parser for F-Deck files.

    Each F-Deck file contains the full history of storm BEST tracks, one storm
    location per line. Example b-deck files are available in the GeoIPS repo.

    Parameters
    ----------
    trackfile_name : str
        Path to fdeck file, with full 6 hourly storm track
        history, formatted as follows:
    allowed_aid_type : list
        List of allowed aid types. Defaults to ["ANAL", "JTWC", "OFCL"] if None

    Returns
    -------
    list
        List of Dictionaries of storm metadata fields from each storm location

    See Also
    --------
    :ref:`api`
        Valid fields can be found in geoips.sector_utils.utils.SECTOR_INFO_ATTRS
    """
    if allowed_aid_types is None:
        allowed_aid_types = ["ANAL", "JTWC", "OFCL"]
    LOG.info("STARTING getting fields from %s", trackfile_name)
    # Must get tcyear out of the filename in case a storm
    # crosses TC vs calendar years.
    # tcyear = os.path.basename(trackfile_name)[5:9]
    tc_year = get_stormyear_from_fdeck_filename(trackfile_name)
    # print tcyear

    flatsf_lines = open(trackfile_name).readlines()
    final_storm_name = get_final_storm_name_fdeck(flatsf_lines, tc_year)
    invest_number = get_invest_number_fdeck(flatsf_lines)

    # This just pulls the time of the first entry in the deck file
    entry_storm_start_datetime = get_storm_start_datetime_from_fdeck_entry(flatsf_lines)

    # Note storms are often started with 3 locations, and the first 2 locations
    # are removed in later deck files, so initial storm start time often does
    # not match the final storm start time.
    # Keep track of the start datetime referenced in the filename
    filename_storm_start_datetime = entry_storm_start_datetime
    # filename_storm_start_datetime = get_storm_start_datetime_from_fdeck_filename( #TLO
    #    trackfile_name  #TLO
    # )  #TLO

    LOG.info("  USING final_storm_name from fdeck %s", final_storm_name)
    LOG.info("  USING storm_start_datetime from fdeck %s", entry_storm_start_datetime)
    LOG.info(
        "  USING original_storm_start_datetime from fdeck %s",
        filename_storm_start_datetime,
    )

    # flatsf_lines go from OLDEST to NEWEST (so firsttime is the OLDEST
    # storm location)
    all_fields = []

    for line in flatsf_lines:
        curr_fields = parse_fdeck_line(
            line,
            source_filename=trackfile_name,
            storm_year=tc_year,
            final_storm_name=final_storm_name,
            invest_number=invest_number,
            storm_start_datetime=entry_storm_start_datetime,
            original_storm_start_datetime=filename_storm_start_datetime,
            parser_name="fdeck_parser",
        )
        # Was previously RE-SETTING finalstormname here.
        # That is why we were getting incorrect final_storm_name fields
        all_fields += [curr_fields]

    LOG.info("FINISHED getting fields from %s", trackfile_name)

    return all_fields, final_storm_name, tc_year, allowed_aid_types


def lat_to_dec(lat_str):
    """Return decimal latitude based on N/S specified string."""
    latnodec = lat_str
    latdec = latnodec[:-2] + "." + latnodec[-2:]
    latdecsign = latdec[:-1] if (latdec[-1] == "N") else "-" + latdec[:-1]
    return latdecsign


def lon_to_dec(lon_str):
    """Return decimal longitude based on E/W specified string."""
    lonnodec = lon_str
    londec = lonnodec[:-2] + "." + lonnodec[-2:]
    londecsign = londec[:-1] if (londec[-1] == "E") else "-" + londec[:-1]
    return londecsign


def parse_fdeck_line(
    line,
    source_filename=None,
    storm_year=None,
    final_storm_name=None,
    invest_number=None,
    storm_start_datetime=None,
    original_storm_start_datetime=None,
    parser_name="fdeck_parser",
):
    """Retrieve the storm information from the current line from the deck file.

    Parameters
    ----------
    line : str
        Current line from the deck file including all storm information
        AL, 99, 2023082312, 03, OFCL,   0, 187N,  651W,  25,    0, TD,  34, NEQ,
        0,    0,    0,    0,    0,    0,   0,  35,   0,    ,   0, ASL, 300,  13,
        AL, 99, 2023082312, 03, OFCL,   3, 188N,  658W,  25, 1009, TD,  34, NEQ,
        0,    0,    0,    0,    0,    0,   0,  35,   0,    ,   0, ASL, 285,  13,
        AL, 99, 2023082312, 03, OFCL,  12, 190N,  678W,  25,    0, LO,  34, NEQ,
        0,    0,    0,    0,    0,    0,   0,  35,   0,    ,   0, ASL, 275,  13,
        AL, 99, 2023082312, 03, OFCL,  24, 191N,  692W,  25,    0, LO,  34, NEQ,
        0,    0,    0,    0,    0,    0,   0,  35,   0,    ,   0, ASL, 270,  16,
        AL, 99, 2023082312, 03, OFCL,  36, 193N,  714W,  25,    0, LO,  34, NEQ,
        0,    0,    0,    0,    0,    0,   0,  35,   0,    ,   0, ASL, 270,  16,

    Returns
    -------
    dict
        Dictionary of the fields from the current storm location from the
        deck file

    See Also
    --------
    :ref:`api`
        Valid fields can be found in geoips.sector_utils.utils.SECTOR_INFO_ATTRS
    """
    # This works with G (GeoIPS) Deck files.
    # Need separate parser for F decks (forecast files)
    # parts = line.split(',', 40)
    parts = [part.strip() for part in line.split(",")]
    if len(parts) != 18 and len(parts) != 23 and len(parts) != 28 and len(parts) != 36:
        raise ValueError(
            "Incorrectly formatted f-deck file - must have either 28 or 36 fields"
        )
    # if len(parts) != 38 and len(parts) != 30 and len(parts) != 40 and
    # len(parts) != 42:
    #    raise ValueError(
    #        "Incorrectly formatted deck file - must have either 30 or 38 or 42 fields"
    #    )
    fields = {}
    fields["deck_line"] = line.strip()
    fields["storm_basin"] = parts[0]
    fields["storm_num"] = int(parts[1])
    fdatetime = datetime.strptime(parts[2], "%Y%m%d%H")
    fields["aid_type"] = parts[
        4
    ]  # BEST, MBAM, OFCL, JTWC, etc - BEST202101220600 when updated
    # CARQ - not best track, real time, A-deck (Aids), F (Fix), E (Error), B (Best)
    fcsttime = int(parts[5])
    fields["synoptic_time"] = fdatetime + timedelta(hours=fcsttime)
    # fields["synoptic_time"] = datetime.strptime(parts[2], "%Y%m%d%H")

    if isinstance(storm_start_datetime, datetime):
        fields["storm_start_datetime"] = storm_start_datetime
    if isinstance(original_storm_start_datetime, datetime):
        fields["original_storm_start_datetime"] = original_storm_start_datetime

    fields["aid_type"] = parts[
        4
    ]  # BEST, MBAM, OFCL, JTWC, etc - BEST202101220600 when updated
    # CARQ - not best track, real time, A-deck (Aids), F (Fix), E (Error), B (Best)
    fields["clat"] = float(lat_to_dec(parts[6]))
    fields["clon"] = float(lon_to_dec(parts[7]))
    fields["vmax"] = parts[8]
    if fields["vmax"]:
        fields["vmax"] = float(fields["vmax"])
    fields["pressure"] = parts[9]
    if fields["pressure"]:
        fields["pressure"] = float(fields["pressure"])

    fields["storm_name"] = "STORM" + parts[1]  # TLO
    # fields["storm_name"] = parts[27] #TLO
    fields["final_storm_name"] = "unknown"
    fields["invest_storm_name"] = "unknown"  # TLO

    storm_id = parts[2].split(" ")[0]  # TLO
    storm_year = storm_id[0:4]  # TLO
    storm_basin = parts[0]  # TLO
    storm_num = parts[1]  # TLO
    fields["storm_year"] = storm_year  # TLO

    if int(parts[1]) > 89:  # TLO
        fields["final_storm_id"] = (
            "tc" + storm_year + storm_basin + storm_num + "invest"
        )  # TLO
        fields["invest_storm_name"] = "INVEST" + storm_num  # TLO
        fields["invest_number"] = storm_num  # TLO
    else:  # TLO
        fields["final_storm_id"] = "tc" + storm_year + storm_basin + storm_num  # TLO
        fields["final_storm_name"] = "STORM" + storm_num  # TLO

    if source_filename:
        from geoips.geoips_utils import replace_geoips_paths

        if isinstance(source_filename, str):
            fields["source_filename"] = replace_geoips_paths(source_filename)
        else:
            fields["source_filename"] = source_filename
    fields["parser_name"] = parser_name
    print("fields=", fields)
    return fields


def get_invest_number_fdeck(deck_lines):
    """Get invest number from full fdeck file."""
    invest_number = None
    for line in deck_lines:
        fields = parse_fdeck_line(line)
        if fields["storm_num"] > 89:
            invest_number = fields["storm_num"]
    return invest_number


def get_storm_start_datetime_from_fdeck_entry(deck_lines):
    """Get storm start datetime from full fdeck file."""
    # Return the synoptic time of the first fdeck entry
    fields = parse_fdeck_line(deck_lines[0])
    LOG.info("  GETTING storm start time from fdeck entry %s", fields["synoptic_time"])
    return fields["synoptic_time"]


def get_storm_start_datetime_from_fdeck_filename(fdeck_filename):
    """Get storm start datetime from fdeck file name."""
    # Return the synoptic time found in the actual filename, if it exists!
    # This will ONLY be the case for INVESTS, which can use the start
    # Gwp912022.2022101400.dat
    fdeck_parts = os.path.basename(fdeck_filename).split(".")
    storm_start_datetime = None
    if len(fdeck_parts) > 2:
        try:
            storm_start_datetime = datetime.strptime(fdeck_parts[1], "%Y%m%d%H")
            LOG.info(
                "  USING storm start time found in filename %s", storm_start_datetime
            )
        except ValueError:
            LOG.warning(
                "  SKIPPING no valid storm start time found in filename, %s",
                "using first entry in fdeck",
            )
            storm_start_datetime = None
    return storm_start_datetime


def get_stormyear_from_fdeck_filename(fdeck_filename):
    """Get the storm year from the F-deck filename.

    Parameters
    ----------
    fdeck_filename : str
        * Path to deck file to search for storm year
        * Must be of format: xxxxxYYYY.dat - pulls YYYY from filename based on
          location

    Returns
    -------
    int
        Storm year
    """
    return int(os.path.basename(fdeck_filename)[5:9])


def get_final_storm_name_fdeck(deck_lines, tcyear):
    """Get final storm name from full fdeck file."""
    final_storm_name = "INVEST"
    for line in deck_lines:
        # curr_fields = parse_fdeck_line(line, tcyear, finalstormname)
        curr_fields = parse_fdeck_line(line, tcyear, final_storm_name)
        # if curr_fields['storm_name']:
        if curr_fields["storm_name"] and curr_fields["storm_name"] != "INVEST":
            LOG.debug("UPDATING final_storm_name to %s", curr_fields["storm_name"])
            final_storm_name = curr_fields["storm_name"]
    return final_storm_name



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Testing module for Pydantic PluginModels."""

from copy import deepcopy

import pytest

from tests.unit_tests.pydantic_models.v1.utils import (
    PathDict,
    load_test_cases,
    load_geoips_yaml_plugin,
    retrieve_model,
    validate_bad_plugin,
    validate_base_plugin,
    validate_neutral_plugin,
)


# A mapping of interfaces implemented in pydantic and a plugin to validate against.
models_available = {
    "feature_annotators": "default_oldlace",
    "gridline_annotators": "default_palegreen",
    "sectors": "korea",
}


def load_good_plugins(models_available):
    """Generate a dictionary of valid GeoIPS plugins.

    Parameters
    ----------
    models_available: dict
        - A dictionary of interfaces and plugin names. Formatted:
          {interface_name: plugin_name}

    Returns
    -------
    good_plugins: dict
        - A dictionary of interfaces and their corresponding valid plugin. Formatted:
          {interface_name: plugin[dict]}
    """
    good_plugins = {}
    for interface, plugin in models_available.items():
        good_plugins[interface] = load_geoips_yaml_plugin(interface, plugin)

    return good_plugins


# Generate separate test cases for each interface
def generate_test_cases(test_type):
    """Generate test cases used for pydantic validation.

    Where each case is a pytest parameter set, Formatted:
    param(argval=(interface_name, test_case[dict]), id={interface_name}_{test_id})

    Parameters
    ----------
    test_type: str
        - The type of test being ran. Must be one of ["bad", "neutral"].

    Returns
    -------
    cases: list[pytest.param]
        - A list of pytest parameters representing cases to test.

    Raises
    ------
    ValueError:
        - Raised if test_type isn't one of 'bad' or 'neutral'. (Raised via
          load_test_cases).
    """
    cases = []
    for interface in models_available.keys():
        test_cases = load_test_cases(interface, test_type)
        for key, value in test_cases.items():
            print(interface, key)
            cases.append(pytest.param((interface, value), id=f"{interface}_{key}"))
    return cases


good_plugins = load_good_plugins(models_available)


@pytest.mark.parametrize(
    "good_plugin",
    good_plugins.values(),
    ids=list(models_available.keys()),
)
def test_good_plugin(good_plugin):
    """Assert that a well formatted GeoIPS plugin is valid.

    Parameters
    ----------
    good_plugin: dict
        - A dictionary representing a valid GeoIPS plugin.
    """
    model = retrieve_model(good_plugin)
    validate_base_plugin(good_plugin, model)


@pytest.mark.parametrize("test_tup", generate_test_cases("bad"))
def test_bad_plugins(test_tup):
    """Perform validation on GeoIPS plugins, including failing cases.

    Parameters
    ----------
    test_tup:
        - A tuple formatted (interface_name, (key, value, class, err_str)), Formatted:
          (str, (str, any, str, str)) used to run and validate tests.
    """
    plugin = PathDict(deepcopy(good_plugins[test_tup[0]]))
    model = retrieve_model(plugin)
    validate_bad_plugin(plugin, test_tup[1], model)


@pytest.mark.parametrize("test_tup", generate_test_cases("neutral"))
def test_neutral_plugins(test_tup):
    """Perform validation on GeoIPS plugins, including neutral cases.

    Parameters
    ----------
    test_tup:
        - A tuple formatted (interface_name, (key, value, class, err_str)), Formatted:
          (str, (str, any, str, str)) used to run and validate tests.
    """
    plugin = PathDict(deepcopy(good_plugins[test_tup[0]]))
    model = retrieve_model(plugin)
    validate_neutral_plugin(plugin, test_tup[1], model)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Standard xarray-based NetCDF output format."""

import logging
from datetime import datetime
import json
import numpy as np

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "xarray_data"
name = "netcdf_xarray"


def call(
    xarray_obj,
    product_names,
    output_fnames,
    clobber=False,
    use_compression=True,
    compression_kwargs=None,
):
    """Write xarray-based NetCDF outputs to disk."""
    for ncdf_fname in output_fnames:
        write_xarray_netcdf(
            xarray_obj,
            ncdf_fname,
            clobber=clobber,
            use_compression=use_compression,
            compression_kwargs=compression_kwargs,
        )
    return output_fnames


def clean_attr_for_netcdf(xobj, attr):
    """Check xarray attributes."""
    # datetime
    if isinstance(xobj.attrs[attr], datetime):
        xobj.attrs[attr] = xobj.attrs[attr].strftime("%c")
    # None cast as string.
    elif xobj.attrs[attr] is None:
        xobj.attrs[attr] = str(xobj.attrs[attr])
    # bools cast as string.
    elif isinstance(xobj.attrs[attr], bool):
        xobj.attrs[attr] = str(xobj.attrs[attr])
    # use json.dumps for dict, list, and tuples.
    elif isinstance(xobj.attrs[attr], (dict, list, tuple)):
        xobj.attrs[attr] = json.dumps(xobj.attrs[attr], default=make_json_friendly)
    # str, bytes, int, float are natively handled
    elif isinstance(xobj.attrs[attr], (str, bytes, int, float)):
        xobj.attrs[attr] = xobj.attrs[attr]
    # other non-native types can just be cast to string.
    # We may want to remove this case, if we want to explicitly handle non-supported
    # types, for easier conversion when reading back in.
    elif not isinstance(xobj.attrs[attr], (str, bytes, int, float)):
        xobj.attrs[attr] = str(xobj.attrs[attr])
    else:
        LOG.warning(
            "SKIPPING attr %s %s, unsupported type %s",
            attr,
            xobj.attrs[attr],
            type(attr),
        )
        xobj.attrs.pop(attr)


def make_json_friendly(value):
    """Return a JSON-serializable version of `value`."""
    # Handle NumPy scalars
    if isinstance(value, np.generic):
        return value.item()  # convert np.int32 -> int, np.float64 -> float, etc.

    # Handle common non-serializable types
    if isinstance(value, (dict, list, tuple)):
        # Recursively make contents JSON-friendly
        return json.loads(json.dumps(value, default=make_json_friendly))
    if isinstance(value, datetime):
        return value.isoformat()

    # Simple types that JSON already understands
    if isinstance(value, (str, int, float, bool)) or value is None:
        return value

    # Fallback: string representation
    return str(value)


def write_xarray_netcdf(
    xarray_obj,
    ncdf_fname,
    clobber=False,
    use_compression=True,
    compression_kwargs=None,
):
    """Write out xarray_obj to netcdf file named ncdf_fname."""
    from geoips.filenames.base_paths import make_dirs
    from os.path import dirname

    make_dirs(dirname(ncdf_fname))

    orig_attrs = xarray_obj.attrs.copy()
    orig_var_attrs = {}

    # Specially handled attributes.
    area_def_str = "none"
    # GEOIPS 1 COMPATIBILITY
    if "area_def" in xarray_obj.attrs.keys():
        # Must pop off the actual area_defintion - does not write to xarray properly
        area_def = xarray_obj.attrs.pop("area_def")
        area_def_str = repr(area_def)
        xarray_obj.attrs["area_def_str"] = area_def_str
    # If actual area_definition object, write it out to xarray as str
    elif "area_definition" in xarray_obj.attrs.keys():
        # If area_definition_str was explicitly defined on the area_definition
        # object, use that
        if hasattr(xarray_obj.area_definition, "area_definition_str"):
            # Must pop off the actual area_defintion - does not write to xarray properly
            area_def_str = xarray_obj.area_definition.area_definition_str
            area_def = xarray_obj.attrs.pop("area_definition")
            xarray_obj.attrs["area_definition_str"] = area_def_str
        else:
            # Must pop off the actual area_defintion - does not write to xarray properly
            area_def = xarray_obj.attrs.pop("area_definition")
            area_def_str = repr(area_def)
            xarray_obj.attrs["area_definition_str"] = area_def_str

    # Standard attribute cleaning for proper serialization for xarray.to_netcdf.
    for attr in xarray_obj.attrs.copy().keys():
        clean_attr_for_netcdf(xarray_obj, attr)
    for varname in xarray_obj.variables.keys():
        orig_var_attrs[varname] = xarray_obj[varname].attrs.copy()
        for attr in xarray_obj[varname].attrs.keys():
            clean_attr_for_netcdf(xarray_obj[varname], attr)

    # Strings to print to the log statement, not used in any other way.
    # If an attribute is not defined, print 'none'.
    roi_str = "none"
    if "interpolation_radius_of_influence" in xarray_obj.attrs.keys():
        roi_str = xarray_obj.interpolation_radius_of_influence

    sdt_str = "none"
    if "start_datetime" in xarray_obj.attrs.keys():
        sdt_str = xarray_obj.attrs["start_datetime"]

    edt_str = "none"
    if "end_datetime" in xarray_obj.attrs.keys():
        edt_str = xarray_obj.attrs["end_datetime"]

    dp_str = "none"
    if "data_provider" in xarray_obj.attrs.keys():
        dp_str = xarray_obj.attrs["data_provider"]

    LOG.info(
        "Writing xarray obj to file %s, source %s, platform %s, start_dt %s, end_dt %s,"
        " %s %s, %s %s, %s %s",
        ncdf_fname,
        xarray_obj.source_name,
        xarray_obj.platform_name,
        sdt_str,
        edt_str,
        "provider",
        dp_str,
        "roi",
        roi_str,
        "area_def",
        area_def_str,
    )

    from os.path import exists

    if use_compression:
        if compression_kwargs is None:
            compression_kwargs = {"zlib": True, "complevel": 5}
        encoding = {x: compression_kwargs for x in list(xarray_obj)}
    else:
        encoding = {}

    # Only re-write the file if requested.
    if clobber is True or not exists(ncdf_fname):
        xarray_obj.to_netcdf(ncdf_fname, encoding=encoding)
    else:
        LOG.warning("SKIPPING not outputing file %s, exists", ncdf_fname)

    # Put the original attributes back at both the dataset level and the variable
    # level. We do not want the serializable attributes on the original dataset.
    xarray_obj.attrs = orig_attrs
    for varname in xarray_obj.variables.keys():
        xarray_obj[varname].attrs = orig_var_attrs[varname]

    return [ncdf_fname]



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""VIIRS SDR Satpy reader.

This VIIRS reader is designed for reading the NPP/JPSS SDR HDF5 files.
The input files are produced by CSPP Polar (CSPP RDR pipeline),
and the read by satpy.

V1.1.0:  NRL-Monterey, Aug. 2024

"""

# Python Standard Libraries
import logging
from pathlib import Path
import os
import re

# Third-Party Libraries
import h5py
import numpy as np
from pandas import DataFrame, date_range, to_datetime
from scipy.interpolate import NearestNDInterpolator
import xarray as xr

# GeoIPS Libraries
from geoips.filenames.base_paths import PATHS as gpaths
from geoips.plugins.modules.readers.utils.geostationary_geolocation import get_indexes
from geoips.utils.context_managers import import_optional_dependencies

# If this reader is not installed on the system, don't fail altogether, just skip this
# import. This reader will not work if the import fails, and the package will have to be
# installed to process data of this type.

LOG = logging.getLogger(__name__)

interface = "readers"
family = "standard"
name = "viirs_sdr_satpy_hdf5"
source_names = ["viirs"]

with import_optional_dependencies(loglevel="info"):
    """Attempt to import a package and print to LOG.info if the import fails."""
    import satpy

VARLIST = {
    "DNB": ["DNB"],
    "IMG": ["I04", "I05"],
    "IMG-Vis": ["I01", "I02", "I03"],
    "MOD": ["M07", "M08", "M10", "M11", "M12", "M13", "M14", "M15", "M16"],
    "MOD-Vis": ["M01", "M02", "M03", "M04", "M05", "M06", "M09"],
}


def bowtie_correction(band, lat, lon):
    """Correct input data for the instrument bowtie effect.

    Correction derived from: Remote Sens. 2016, 8, 79.
    """
    # Unfold
    ord_lat = np.sort(lat, axis=0)
    unfold_idx = np.argsort(lat, axis=0)

    rad_fold = np.take_along_axis(band, unfold_idx, axis=0)
    sort_lon = np.take_along_axis(lon, unfold_idx, axis=0)

    if np.all(np.isnan(rad_fold)):
        LOG.debug("All nan band, no bowtie correction")
        return rad_fold, ord_lat, sort_lon
    if np.all(np.isnan(rad_fold)):
        LOG.debug("All NaNs, no bowtie correction")
        return rad_fold, ord_lat, sort_lon

    res_band = rad_fold.copy()
    # mask and indicies of bad data
    rad_nan_flag = np.isnan(rad_fold)
    nan_mask = np.where(~rad_nan_flag)

    interp = NearestNDInterpolator(np.transpose(nan_mask), rad_fold[nan_mask])
    interp_nans = interp(*np.indices(rad_fold.shape)[:, rad_nan_flag], **{"workers": 4})
    res_band[rad_nan_flag] = interp_nans

    return res_band, ord_lat.astype(np.float64), sort_lon.astype(np.float64)


def check_valid_satpy_filename(filename):
    """Check if VIIRS SDR filename is supported by satpy reader.

    VIIRS SDR HDF5 files must include the creation time in filename after the start and
    end times, otherwise will not recognize the file.

    Parameters
    ----------
    filename : str
        Filename of the VIIRS SDR HDF5 file (basename, not full file path)

    Returns
    -------
    bool
        True if valid satpy filename
    """
    complete_filename_pattern = (
        r"^[A-Za-z0-9]{5}_[a-z]{1}\d{2}_d\d{8}_t\d{7}_e\d{7}_b\d{5}_c\d{15,21}"
        r"_[a-zA-Z0-9_]{5,10}\.h5$"
    )
    return bool(re.match(complete_filename_pattern, filename))


def check_filename_missing_ctime(filename):
    """Check if VIIRS SDR filename is missing the creation time.

    VIIRS SDR HDF5 files must include the creation time in filename after the start and
    end times, otherwise will not recognize the file. If this function returns True,
    the filename will be supported by satpy if the creation time is added to the name.

    Parameters
    ----------
    filename : str
        Filename of the VIIRS SDR HDF5 file (basename, not full file path)

    Returns
    -------
    bool
        True if the filename contains the start and end times, but is missing the ctime
    """
    missing_ctime_filename_pattern = (
        r"^[A-Za-z0-9]{5}_[a-z]{1}\d{2}_d\d{8}_t\d{7}_e\d{7}_b\d{5}_[a-zA-Z0-9_]{5,10}"
        r"\.h5$"
    )
    return bool(re.match(missing_ctime_filename_pattern, filename))


def add_creation_time_to_filename(filename):
    """Add a product creation time to an SDR file that is missing the creation time.

    We will use the end time in the filename as a placeholder for the creation time.

    Parameters
    ----------
    filename : str
        Filename of the VIIRS SDR HDF5 file (basename, not full file path)

    Returns
    -------
    str
        Filename with creation time inserted after d%Y%m%d_s%H%M%S_e%H%MS_bNNNNN.
    """
    split_fname = filename.split("_")
    creation_time = "c" + split_fname[2][1:] + split_fname[4][1:]
    return "_".join(split_fname[:-2] + [creation_time] + split_fname[-2:])


def check_filenames_and_link(filepaths):
    """Attempt to link to cache directory with valid name if applicable/able.

    Parameters
    ----------
    filepaths : list
        Filepaths passed to reader

    Returns
    -------
    dict
        Lists of valid files, linked files, and invalid files
    """
    target_link_dir = os.path.join(
        gpaths["GEOIPS_DATA_CACHE_DIR_SHORTTERM_CALIBRATED_DATA"],
        "viirs_sdr_hdf5_reader",
    )
    checked = {"valid": [], "linked": [], "invalid": []}
    for filepath in filepaths:
        filename = Path(filepath).name
        if check_valid_satpy_filename(filename):
            # This filename is supported by satpy - no action required
            checked["valid"].append(filepath)
        elif check_filename_missing_ctime(filename):
            # This is a valid VIIRS SDR file, but is missing the creation time
            # Use the end time as the creation time, and link to GeoIPS cache dir
            new_filename = add_creation_time_to_filename(filename)
            if not Path(target_link_dir).exists():
                Path(target_link_dir).mkdir(parents=True, exist_ok=True)
            linked_path = os.path.join(target_link_dir, new_filename)
            if not Path(linked_path).exists():
                LOG.info("Linking %s as %s", filepath, linked_path)
                Path(linked_path).symlink_to(filepath)
            checked["linked"].append(linked_path)
        else:
            # This filename is right out
            checked["invalid"].append(filepath)
    return checked


def call(
    fnames,
    metadata_only=False,
    chans=None,
    area_def=None,
    self_register=False,
    resample=False,
    bowtie=True,
):
    """Read VIIRS SDR hdf5 data products.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.
        * Conforms to geoips xarray standards, see more in geoips documentation.
    """
    # print("Reading")
    check_fnames = check_filenames_and_link(fnames)
    fnames = check_fnames["valid"] + check_fnames["linked"]
    tmp_scn = satpy.Scene(reader="viirs_sdr", filenames=fnames)
    scn_start, scn_end = tmp_scn.start_time, tmp_scn.end_time
    base_fnames = list(map(os.path.basename, fnames))

    full_xr = {}
    if metadata_only:
        # average resolution
        # sensor, plat name
        tmp_scn.load([tmp_scn.available_dataset_names()[0]])
        tmp_attrs = tmp_scn[tmp_scn.available_dataset_names()[0]].attrs
        tmp_xr = xr.Dataset(
            attrs={
                "source_file_name": base_fnames[0],
                "start_datetime": scn_start,
                "end_datetime": scn_end,
                "source_name": tmp_attrs["sensor"],
                "platform_name": tmp_attrs["platform_name"],
                "data_provider": "NOAA",
                "sample_distance_km": 1,
                "interpolation_radius_of_influence": 1000,  # guess!
            }
        )
        tmp_dict = {"METADATA": tmp_xr}
        for lfile in check_fnames["linked"]:
            if Path(lfile).exists():
                LOG.info("Unlinking %s", lfile)
                Path(lfile).unlink()
        return tmp_dict

    # trim VARLIST based on channels requested
    if chans:
        tmp_vl = VARLIST.copy()
        for key, val in tmp_vl.items():
            km = [c[:3] in val for c in chans]
            if not any(km):
                VARLIST.pop(key)
            else:
                matches = [s for s in val for c in chans if s in c]
                VARLIST[key] = matches

    # could opimize more
    tmp_coor = {}
    for var in VARLIST:
        tmp_dask = {}
        dataset_ids = [
            idx
            for idx in tmp_scn.available_dataset_ids()
            if idx["name"] in VARLIST[var]
        ]
        if len(dataset_ids) == 0:
            LOG.warning("No variables found in data for viirs_sdr reader.")
            continue

        for d in dataset_ids:
            tmp_scn.load([d])
            full_key = tmp_scn[d].attrs["name"] + tmp_scn[d].attrs[
                "calibration"
            ].capitalize()[:3].replace("Bri", "BT")

            tmp_ma = tmp_scn[d].to_masked_array().data

            # load band
            tmp_scn.load([d])

            lat = tmp_scn[d].area.lats.to_masked_array().data
            lon = tmp_scn[d].area.lons.to_masked_array().data

            # bowtie correction
            if bowtie:
                band_data, band_lat, band_lon = bowtie_correction(tmp_ma, lat, lon)
            else:
                band_data, band_lat, band_lon = tmp_ma, lat, lon

            if "Ref" in full_key:
                from geoips.data_manipulations.corrections import apply_gamma

                # gamma expects data 0-1 range
                # Gamma factor derived from log rule
                # log(mean)/log(0.5) for mean rng 0 to 1
                band_data = apply_gamma(band_data / 100, 1.65) * 100

            tmp_dask |= {full_key: (("dim_0", "dim_1"), band_data)}

        # coordinates
        tmp_coor["latitude"] = (
            ("dim_0", "dim_1"),
            band_lat,
        )
        tmp_coor["longitude"] = (
            ("dim_0", "dim_1"),
            band_lon,
        )

        # sample time to the proper shape (N*48), while lat/lon are ()
        time_range = date_range(
            start=scn_start,
            end=scn_end,
            periods=tmp_coor["latitude"][1].shape[0],
        ).values
        interp_time = np.tile(time_range, (tmp_coor["latitude"][1].shape[1], 1)).T
        tmp_coor["time"] = (("dim_0", "dim_1"), interp_time)

        tmp_attrs = tmp_scn[VARLIST[var][0]].attrs

        cal_params = [
            "satellite_azimuth_angle",
            "satellite_zenith_angle",
            "solar_azimuth_angle",
            "solar_zenith_angle",
        ]

        if var == "DNB":
            cal_params = [
                "dnb_lunar_azimuth_angle",
                "dnb_lunar_zenith_angle",
                "dnb_satellite_azimuth_angle",
                "dnb_satellite_zenith_angle",
                "dnb_solar_azimuth_angle",
                "dnb_solar_zenith_angle",
            ]

        tmp_scn.load(cal_params)
        tmp_cal_params = {
            i.removeprefix("dnb_"): (
                ("dim_0", "dim_1"),
                tmp_scn[i].to_masked_array(),
            )
            for i in cal_params
        }
        if var == "DNB":
            # tmp_scn.load(["dnb_moon_illumination_fraction"])
            dnb_geofile = [i for i in fnames if "GDNBO" in os.path.basename(i)][0]
            h5_dnb = h5py.File(dnb_geofile)
            phase_ang = h5_dnb["All_Data/VIIRS-DNB-GEO_All/MoonPhaseAngle"][...]

            tmp_cal_params["moon_frac"] = (
                ("dim_0", "dim_1"),
                np.full(
                    tmp_cal_params[cal_params[0].removeprefix("dnb_")][1].shape,
                    0.5 * (1 + np.cos(phase_ang)),
                ),
            )

        if var == "DNB":
            try:
                from lunarref.lib.liblunarref import lunarref

                # this results in the wrong value..
                # np.arccos((tmp_scn["dnb_moon_illumination_fraction"].data/50)-1)

                lunarref_data = lunarref(
                    tmp_dask["DNBRad"][1],
                    tmp_cal_params["solar_zenith_angle"][1],
                    tmp_cal_params["lunar_zenith_angle"][1],
                    scn_start.strftime("%Y%m%d%H"),
                    scn_start.strftime("%M"),
                    phase_ang,
                )
                lunarref_data = np.ma.masked_less_equal(lunarref_data, -999, copy=False)
                tmp_dask |= {
                    "DNBRef": (("dim_0", "dim_1"), lunarref_data),
                }

            except ImportError:
                LOG.info("Failed lunarref in viirs reader.  If you need it, build it")

        # problem with sat_za/az values being too high, need to downsample
        # print("Building xarray")
        obs_xr = xr.Dataset(data_vars=tmp_dask)
        coor_xr = xr.Dataset(data_vars=tmp_coor)
        cal_xr = xr.Dataset(data_vars=tmp_cal_params)

        if cal_xr.sizes["dim_0"] > coor_xr.sizes["dim_0"]:
            cal_xr = cal_xr.coarsen(dim_0=2, dim_1=2).mean()

        tmp_xr = xr.merge([obs_xr, coor_xr, cal_xr])

        tmp_xr.attrs = {
            "source_file_name": base_fnames,
            "start_datetime": tmp_scn.start_time,
            "end_datetime": tmp_scn.end_time,
            "source_name": tmp_attrs["sensor"],
            "platform_name": tmp_attrs["platform_name"],
            "data_provider": "NOAA",
            "sample_distance_km": tmp_attrs["resolution"] / 1e3,
            "interpolation_radius_of_influence": 1000,
        }

        full_xr |= {var: tmp_xr}
    full_xr["METADATA"] = xr.Dataset(attrs=tmp_xr.attrs)

    # Geolocation resampling
    if resample and area_def:
        adname = area_def.area_id
        new_shape = area_def.shape

        LOG.info("")
        LOG.info("Getting geolocation information for {}.".format(adname))

        for dtype in full_xr.keys():
            if "latitude" not in full_xr[dtype].variables:
                LOG.info(
                    "No data read for dataset %s, removing from xarray list", dtype
                )
                continue
            fldk_lats = full_xr[dtype]["latitude"]
            fldk_lons = full_xr[dtype]["longitude"]
            # Get just the metadata we need
            geo_metadata = full_xr[dtype].attrs.copy()
            geo_metadata["num_lines"], geo_metadata["num_samples"] = fldk_lats.shape
            geo_metadata["scene"] = "stitched_granules"
            geo_metadata["fnames"] = geo_metadata.pop("source_file_name")
            geo_metadata["roi_factor"] = 5
            lines, samples = get_indexes(geo_metadata, fldk_lats, fldk_lons, area_def)

            index_mask = lines != -999

            new_dim0 = "dim_{:d}".format(new_shape[0])
            new_dim1 = "dim_{:d}".format(new_shape[1])

            for varname in full_xr[dtype].variables.keys():
                new_var = np.full(new_shape, -999.1)
                new_var[index_mask] = full_xr[dtype][varname].values[
                    lines[index_mask], samples[index_mask]
                ]
                # out_var = new_var
                if varname not in cal_params + ["latitude", "longitude"]:
                    # Set values <= -999.9 to NaN so they also get interpolated.
                    new_var[np.where(new_var <= -999.9)] = np.nan
                    # Interpolate missing data.
                    out_var = np.array(DataFrame(new_var).interpolate())
                else:
                    out_var = new_var
                full_xr[dtype][varname] = xr.DataArray(
                    out_var, dims=[new_dim0, new_dim1]
                )
            if "time" in full_xr[dtype].variables:
                time_data = full_xr[dtype]["time"].data
                time_data = np.asarray(to_datetime(time_data.ravel())).reshape(
                    time_data.shape
                )
                full_xr[dtype]["time"].data = time_data

            ll_mask = full_xr[dtype]["latitude"].data >= -90
            for varname in full_xr[dtype].variables.keys():
                full_xr[dtype][varname] = full_xr[dtype][varname].where(ll_mask)

        LOG.interactive("Done with geolocation for {}".format(adname))
        LOG.info("")
    for lfile in check_fnames["linked"]:
        if Path(lfile).exists():
            LOG.info("Unlinking %s", lfile)
            Path(lfile).unlink()
    return full_xr



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Pydantic models used to validate GeoIPS gridline annotator plugins."""

from typing import Literal, Union, List, Optional

from pydantic import Field

from geoips.pydantic_models.v1.bases import (
    FrozenModel,
    PermissiveFrozenModel,
    PluginModel,
    ColorType,
)


class Spacing(FrozenModel):
    """Model used to format the spacing of gridlines in annotated imagery."""

    latitude: Union[float, Literal["auto"]] = Field(
        ...,
        description=(
            "Latitude spacing in degrees, can be a float or a string. If a string, it "
            "must be 'auto', which represents automatic spacing based on your area_def."
        ),
    )
    longitude: Union[float, Literal["auto"]] = Field(
        ...,
        description=(
            "Longitude spacing in degrees, can be a float or a string. If a string, it "
            "must be 'auto', which represents automatic spacing based on your area_def."
        ),
    )


class Labels(PermissiveFrozenModel):
    """Model used to format labels in annotated imagery.

    For more information, visit https://matplotlib.org/stable/ to see more context in
    how to specify these fields.
    """

    top: bool = Field(..., description="Whether to display the label at the top.")
    bottom: bool = Field(..., description="Whether to display the label at the bottom.")
    left: bool = Field(..., description="Whether to display the label on the left.")
    right: bool = Field(..., description="Whether to display the label on the right.")
    alpha: Optional[float] = Field(
        None,
        description=(
            "Inverse percentage of label transparency (0 = fully transparent, 1 = fully"
            "opaque)."
        ),
    )
    backgroundcolor: Optional[ColorType] = Field(
        None,
        description=(
            "A rgb tuple, matplotlib named color, or hexidecimal string (#XXXXXX) "
            "to apply to the background of your label frame.\n"
            "For more info, see: "
            "https://matplotlib.org/stable/users/explain/colors/colors.html"
        ),
    )
    color: Optional[ColorType] = Field(
        None,
        description=(
            "A rgb tuple, matplotlib named color, or hexidecimal string (#XXXXXX) "
            "to apply to the label text of your image.\n"
            "For more info, see: "
            "https://matplotlib.org/stable/users/explain/colors/colors.html"
        ),
    )
    fontfamily: Optional[str] = Field(
        None, description="Font family for the label text."
    )
    fontsize: Optional[Union[float, str]] = Field(
        None, description="Font size for the label text."
    )
    fontstretch: Optional[Union[int, str]] = Field(
        None, description="Stretch level of the font."
    )
    fontstyle: Optional[str] = Field(
        None, description="Style of the font (e.g., 'normal', 'italic', 'oblique')."
    )
    fontvariant: Optional[str] = Field(
        None, description="Font variant (e.g., 'normal', 'small-caps')."
    )
    fontweight: Optional[Union[int, str]] = Field(
        None, description="Weight of the font."
    )
    linespacing: Optional[float] = Field(
        None, description="Spacing between lines of text."
    )
    mouseover: Optional[bool] = Field(
        None, description="Whether to enable mouseover effect."
    )
    position: Optional[List[float]] = Field(
        None, description="Absolute (x, y) position of the label."
    )
    rotation: Optional[Union[float, str]] = Field(
        None, description="Rotation of the label (degrees or 'vertical'/'horizontal')."
    )
    rotation_mode: Optional[str] = Field(
        None, description="Rotation mode ('default' or 'anchor')."
    )
    snap: Optional[bool] = Field(
        None, description="Whether to snap the label to the pixel grid."
    )
    wrap: Optional[bool] = Field(None, description="Whether to allow text wrapping.")
    zorder: Optional[float] = Field(
        None, description="Order of label rendering (lower value = earlier rendering)."
    )
    xpadding: Optional[int] = Field(None, description="Pixel offset for x-axis labels.")
    ypadding: Optional[int] = Field(None, description="Pixel offset for y-axis labels.")


class Lines(FrozenModel):
    """Model used to format gridlines in annotated imagery.

    For more information, visit https://matplotlib.org/stable/ to see more context in
    how to specify these fields.
    """

    color: ColorType = Field(
        ..., description="Color of the line (named color, hex, or rgb tuple)."
    )
    linestyle: List[int] = Field(
        ...,
        min_items=2,
        max_items=2,
        description="Pattern of dashes and gaps in the line.",
    )
    linewidth: float = Field(..., description="Width of the line.")
    # TODO: create file called matplotlib or 'artists', which contains all the machinery
    # that relates with this


class GridlineAnnotatorSpec(FrozenModel):
    """Gridline Annotator spec (specification) format."""

    spacing: Spacing = Field(..., description="Spacing settings for the plot.")
    labels: Labels = Field(..., description="Label settings for the plot.")
    lines: Lines = Field(..., description="Line settings for the plot.")
    background: Optional[ColorType] = Field(
        None,
        description=(
            "A rgb tuple, matplotlib named color, or hexidecimal string (#XXXXXX) "
            "to apply to the background of your image frame.\n"
            "For more info, see: "
            "https://matplotlib.org/stable/users/explain/colors/colors.html"
        ),
    )


class GridlineAnnotatorPluginModel(PluginModel):
    """Gridline Annotator plugin format."""

    spec: GridlineAnnotatorSpec = Field(
        ...,
        description=(
            "Specification of how to apply gridlines, labels, and spacing to your "
            "annotated imagery. Works alongside matplotlib and cartopy to generate "
            "these attributes."
        ),
    )



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Arctic Weather Satellite (aws) MWR Level 1B NetCDF reader.

Channel information is from the ESA webpage:
https://www.esa.int/Applications/Observing_the_Earth/Meteorological_missions/

Arctic_Weather_Satellite/The_instrument
All channels have a same array size (nscan,nfov).

Chan        Freq(GHz)        FOV(km)        Utilisation

Band 1:
1           50.3             <= 40          Temp sounding
2           52.8             <= 40          Temp sounding
3           53.246           <= 40          Temp sounding
4           53.956           <= 40          Temp sounding
5           54.4             <= 40          Temp sounding
6           54.94            <= 40          Temp sounding
7           55.5             <= 40          Temp sounding
8           57.290           <= 40          Temp sounding

Band 2:
9           89               <= 20          Window and cloud detection

Band 3:
10          165.5            <= 10          Window and humidity sounding
11          176.31           <= 10          Humidity sounding
12          178.81           <= 10          Humidity sounding
13          180.31           <= 10          Humidity sounding
14          181.51           <= 10          Humidity sounding
15          182.31           <= 10          Humidity sounding

Band 4:
16          325.15 +- 1.2    <= 10          Humidity sounding/cloud detection
17          325.15 +- 2.4    <= 10          Humidity sounding/cloud detection
18          325.15 +- 4.1    <= 10          Humidity sounding/cloud detection
19          325.15 +- 6.6    <= 10          Humidity sounding/cloud detection

Note:  We do not know actual assignments of channels to the four groups, due to lack of
official documentation on Arctic Weather Satellite data. Thus, we select current
channels for each band based on internal discussions and our experience.
We will make adjustments once there are any differences from official document is
available.

Also, variables (L1B_quality_flag and degraded_channels) are space holders for QC
purposes. They are not used for now, but will be applied later.
"""

import logging
from datetime import datetime

# import os
# from collections import OrderedDict
# import numpy as np
import xarray

# from IPython import embed as shell
LOG = logging.getLogger(__name__)

#                            AWS sensor info
# setup variables
ROIs = {"Band1": 45000, "Band2": 25000, "Band3": 20000, "Band4": 20000}
group_names = ["/data/navigation", "/data/calibration", "/quality"]
list_vars = [
    "aws_lat",
    "aws_lon",
    "aws_solar_zenith_angle",
    "aws_solar_azimuth_angle",
    "aws_satellite_zenith_angle",
    "aws_satellite_azimuth_angle",
    "L1B_quality_flag",
    "degraded_channels",
    "aws_toa_brightness_temperature",
]

interface = "readers"
family = "standard"
name = "aws_netcdf"


def merge_xarrays(src_xobj, dst_xobj):
    """Merge one xarray into another.

    Parameters
    ----------
    src_xobj : xarray.Dataset
        Input dataset to merge
    dst_xobj : xarray.Dataset
        Dataset where src_xobj will be merged

    Returns
    -------
    xarray.Dataset
        Merged xarray dataset
    """
    tmp_xobj = xarray.Dataset()
    for dvar in src_xobj.data_vars.keys():
        dset = src_xobj[dvar]
        if dvar not in dst_xobj:
            dst_xobj[dvar] = dset
        else:
            tmp_xobj[dvar] = xarray.concat(
                [dst_xobj[dvar], dset], dim=dst_xobj[dvar].dims[0]
            )
            dst_xobj = dst_xobj.drop(dvar)
    try:
        merged_xobj = xarray.merge([tmp_xobj, dst_xobj])
    except ValueError:
        tmp_dims = {x: tmp_xobj[x].size for x in tmp_xobj.dims}
        dst_dims = {x: dst_xobj[x].size for x in dst_xobj.dims}
        mismatching_dims = [x for x in dst_dims if tmp_dims[x] != dst_dims[x]]
        merged_xobj = xarray.combine_nested(
            [tmp_xobj, dst_xobj], concat_dim=mismatching_dims
        )
    return merged_xobj


def get_geoips_platform_name(meta_platform_name):
    """Map source file platform name to standard GeoIPS names.

    Parameters
    ----------
    meta_platform_name : str
        Name of platform in source file

    Returns
    -------
    str
        Standard GeoIPS platform name
    """
    geoips_pform_name = {"AWS1": "aws"}
    return geoips_pform_name[meta_platform_name]


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read aws L1 netCDF data.

    Parameters
    ----------
    fci_files : list
        list of files for a given scan time
    metadata_only : bool, optional
        Only return metadata, by default False
    chans : list, optional
        List of channels/variables to load, by default None
    area_def : pyresample area definition, optional
        Read in data for specific area definition, by default None
        (currently unsupported)
    self_register : bool, optional
        Self register data to given resolution, by default False
        (currently unsupported)

    Returns
    -------
    dict
        dictionary of xarray datasets

    Raises
    ------
    AttributeError
        Encountered unknown platform name in source file

    Notes
    -----
    data : grouped file
        stored in groups (status, data, quality) and sub-groups
        open_groups to open and decode a file for required fields
        global attributes in home group '/'
        '/'  is an important character to access data
        process only one orbit file
    """
    xobj_group = xarray.Dataset()
    for fname in fnames:  # loop input files (only one for now)
        LOG.info(f"Opening: {fname}")
        # First grab metadata
        ds = xarray.open_groups(fname)
        available_groups = list(ds)

        meta = xarray.Dataset(attrs=ds["/"].attrs)
        try:
            if "." in meta.attrs["sensing_start_time_utc"]:
                tformat = "%Y-%m-%d %H:%M:%S.%f"
            else:
                tformat = "%Y-%m-%d %H:%M:%S"
            start_datetime = datetime.strptime(
                meta.attrs["sensing_start_time_utc"], tformat
            )
            end_datetime = datetime.strptime(
                meta.attrs["sensing_end_time_utc"], tformat
            )
        except KeyError:
            LOG.info("error in orbit data start and end time")
        if hasattr(meta, "Spacecraft"):
            pform = meta.attrs["Spacecraft"]
        else:
            raise AttributeError("Cannot identify platform name")
        geoips_attrs = {
            "area_definition": area_def,
            "start_datetime": start_datetime,
            "end_datetime": end_datetime,
            "vertical_data_type": "surface",
            "source_name": "mwr",
            "platform_name": get_geoips_platform_name(pform),
            "data_provider": "noaa",
            "interpolation_radius_of_influence": 20000,
        }
        meta.attrs = dict(meta.attrs, **geoips_attrs)

        if metadata_only is True:
            LOG.info(
                "metadata_only is True, reading only first"
                + "file for metadata information and returning"
            )
            return {"METADATA": meta}
        # load data for list_vars from these groups
        for group in group_names:
            if group in available_groups:
                available_vars_group = list(ds[group].data_vars.keys())
                # Only load the requested variables from this group
                drop_vars = list(set(available_vars_group).difference(set(list_vars)))
                xr = xarray.load_dataset(fname, group=group, drop_variables=drop_vars)
            # merge xarray
            xobj_group = merge_xarrays(xr, xobj_group)

    # setup 4-band xarrays
    # Channel assignments for each band are based on internal discussions and
    # our experience.  They could be adjusted later once data documentations are
    # available.
    # quality_flag and degraded_channels are spare variables for QC purposes. They
    # are not used now, but will be applied later.
    xobj_band1 = xarray.Dataset()
    xobj_band2 = xarray.Dataset()
    xobj_band3 = xarray.Dataset()
    xobj_band4 = xarray.Dataset()

    xobj_band1["chan1"] = xobj_group["aws_toa_brightness_temperature"][:, :, 0]
    xobj_band1["chan2"] = xobj_group["aws_toa_brightness_temperature"][:, :, 1]
    xobj_band1["chan3"] = xobj_group["aws_toa_brightness_temperature"][:, :, 2]
    xobj_band1["chan4"] = xobj_group["aws_toa_brightness_temperature"][:, :, 3]
    xobj_band1["chan5"] = xobj_group["aws_toa_brightness_temperature"][:, :, 4]
    xobj_band1["chan6"] = xobj_group["aws_toa_brightness_temperature"][:, :, 5]
    xobj_band1["chan7"] = xobj_group["aws_toa_brightness_temperature"][:, :, 6]
    xobj_band1["chan8"] = xobj_group["aws_toa_brightness_temperature"][:, :, 7]
    xobj_band2["chan9"] = xobj_group["aws_toa_brightness_temperature"][:, :, 8]
    xobj_band3["chan10"] = xobj_group["aws_toa_brightness_temperature"][:, :, 9]
    xobj_band3["chan11"] = xobj_group["aws_toa_brightness_temperature"][:, :, 10]
    xobj_band3["chan12"] = xobj_group["aws_toa_brightness_temperature"][:, :, 11]
    xobj_band3["chan13"] = xobj_group["aws_toa_brightness_temperature"][:, :, 12]
    xobj_band3["chan14"] = xobj_group["aws_toa_brightness_temperature"][:, :, 13]
    xobj_band3["chan15"] = xobj_group["aws_toa_brightness_temperature"][:, :, 14]
    xobj_band4["chan16"] = xobj_group["aws_toa_brightness_temperature"][:, :, 15]
    xobj_band4["chan17"] = xobj_group["aws_toa_brightness_temperature"][:, :, 16]
    xobj_band4["chan18"] = xobj_group["aws_toa_brightness_temperature"][:, :, 17]
    xobj_band4["chan19"] = xobj_group["aws_toa_brightness_temperature"][:, :, 18]

    xobj_band1["latitude"] = xobj_group["aws_lat"][:, :, 0]
    xobj_band1["latitude"] = xobj_group["aws_lat"][:, :, 0]
    xobj_band1["longitude"] = xobj_group["aws_lon"][:, :, 0]
    xobj_band1["solar_zenith_angle"] = xobj_group["aws_solar_zenith_angle"][:, :, 0]
    xobj_band1["solar_azimuth_angle"] = xobj_group["aws_solar_azimuth_angle"][:, :, 0]
    xobj_band1["satellite_zenith_angle"] = xobj_group["aws_satellite_zenith_angle"][
        :, :, 0
    ]
    xobj_band1["satellite_azimuth_angle"] = xobj_group["aws_satellite_azimuth_angle"][
        :, :, 0
    ]
    xobj_band1["quality_flag"] = xobj_group["L1B_quality_flag"]
    xobj_band1["degraded_channels"] = xobj_group["degraded_channels"]

    xobj_band2["latitude"] = xobj_group["aws_lat"][:, :, 1]
    xobj_band2["latitude"] = xobj_group["aws_lat"][:, :, 1]
    xobj_band2["longitude"] = xobj_group["aws_lon"][:, :, 1]
    xobj_band2["solar_zenith_angle"] = xobj_group["aws_solar_zenith_angle"][:, :, 1]
    xobj_band2["solar_azimuth_angle"] = xobj_group["aws_solar_azimuth_angle"][:, :, 1]
    xobj_band2["satellite_zenith_angle"] = xobj_group["aws_satellite_zenith_angle"][
        :, :, 1
    ]
    xobj_band2["satellite_azimuth_angle"] = xobj_group["aws_satellite_azimuth_angle"][
        :, :, 1
    ]
    xobj_band2["quality_flag"] = xobj_group["L1B_quality_flag"]
    xobj_band2["degraded_channels"] = xobj_group["degraded_channels"]

    xobj_band3["latitude"] = xobj_group["aws_lat"][:, :, 2]
    xobj_band3["latitude"] = xobj_group["aws_lat"][:, :, 2]
    xobj_band3["longitude"] = xobj_group["aws_lon"][:, :, 2]
    xobj_band3["solar_zenith_angle"] = xobj_group["aws_solar_zenith_angle"][:, :, 2]
    xobj_band3["solar_azimuth_angle"] = xobj_group["aws_solar_azimuth_angle"][:, :, 2]
    xobj_band3["satellite_zenith_angle"] = xobj_group["aws_satellite_zenith_angle"][
        :, :, 2
    ]
    xobj_band3["satellite_azimuth_angle"] = xobj_group["aws_satellite_azimuth_angle"][
        :, :, 2
    ]
    xobj_band3["quality_flag"] = xobj_group["L1B_quality_flag"]
    xobj_band3["degraded_channels"] = xobj_group["degraded_channels"]

    xobj_band4["latitude"] = xobj_group["aws_lat"][:, :, 3]
    xobj_band4["latitude"] = xobj_group["aws_lat"][:, :, 3]
    xobj_band4["longitude"] = xobj_group["aws_lon"][:, :, 3]
    xobj_band4["solar_zenith_angle"] = xobj_group["aws_solar_zenith_angle"][:, :, 3]
    xobj_band4["solar_azimuth_angle"] = xobj_group["aws_solar_azimuth_angle"][:, :, 3]
    xobj_band4["satellite_zenith_angle"] = xobj_group["aws_satellite_zenith_angle"][
        :, :, 3
    ]
    xobj_band4["satellite_azimuth_angle"] = xobj_group["aws_satellite_azimuth_angle"][
        :, :, 3
    ]
    xobj_band4["quality_flag"] = xobj_group["L1B_quality_flag"]
    xobj_band4["degraded_channels"] = xobj_group["degraded_channels"]

    # add attributes to each band (will make adjustments on roi etc later)
    geoips_attrs["interpolation_radius_of_influence"] = ROIs["Band1"]
    xobj_band1.attrs = geoips_attrs.copy()
    geoips_attrs["interpolation_radius_of_influence"] = ROIs["Band2"]
    xobj_band2.attrs = geoips_attrs.copy()
    geoips_attrs["interpolation_radius_of_influence"] = ROIs["Band3"]
    xobj_band3.attrs = geoips_attrs.copy()
    geoips_attrs["interpolation_radius_of_influence"] = ROIs["Band4"]
    xobj_band4.attrs = geoips_attrs.copy()

    return {
        "AWS_Band1": xobj_band1,
        "AWS_Band2": xobj_band2,
        "AWS_Band3": xobj_band3,
        "AWS_Band4": xobj_band4,
        "METADATA": meta,
    }



"""Test script for representative product comparisons using GeoTIFF files."""

from __future__ import annotations

import logging
from functools import partial
from pathlib import Path
from typing import Sequence

import numpy as np
import rasterio
from rasterio.errors import RasterioIOError

from geoips.commandline.log_setup import log_with_emphasis

LOG = logging.getLogger(__name__)

interface = "output_checkers"
family = "standard"
name = "geotiff"

# Comparison configuration
COMPARISON_TOLERANCE = 1e-6
VALID_EXTENSIONS = frozenset({".tif", ".tiff", ".geotif", ".geotiff"})
IGNORED_METADATA_FIELDS = frozenset({"creation_timestamp", "datetime", "created"})
RANDOM_SEED = 42

# Test noise scales
CLOSE_MISMATCH_NOISE_SCALE = 0.05
BAD_MISMATCH_NOISE_SCALE = 0.25


def correct_file_format(fname: str) -> bool:
    """Determine if fname is a GeoTIFF file.

    Parameters
    ----------
    fname : str
        Name of file to check.

    Returns
    -------
    bool
        True if it is a GeoTIFF file, False otherwise.
    """
    return Path(fname).suffix.lower() in VALID_EXTENSIONS


def _validate_file_exists(filepath: Path) -> bool:
    """Validate that a file exists at the given path.

    Parameters
    ----------
    filepath : Path
        Path to validate.

    Returns
    -------
    bool
        True if file exists, False otherwise.
    """
    if not filepath.exists():
        LOG.error("File not found: %s", filepath)
        return False
    return True


def _filter_metadata(profile: dict) -> dict:
    """Filter out ignored fields from rasterio profile metadata.

    Parameters
    ----------
    profile : dict
        Rasterio profile dictionary.

    Returns
    -------
    dict
        Filtered profile with ignored fields removed.
    """
    return {
        key: value
        for key, value in profile.items()
        if key.lower() not in IGNORED_METADATA_FIELDS
    }


def _compare_metadata(output_profile: dict, compare_profile: dict) -> bool:
    """Compare two rasterio profiles, ignoring specified fields.

    Parameters
    ----------
    output_profile : dict
        Profile from output product.
    compare_profile : dict
        Profile from comparison product.

    Returns
    -------
    bool
        True if profiles match, False otherwise.
    """
    filtered_output = _filter_metadata(output_profile)
    filtered_compare = _filter_metadata(compare_profile)
    return filtered_output == filtered_compare


def _compare_arrays(output_data: np.ndarray, compare_data: np.ndarray) -> bool:
    """Compare two numpy arrays with appropriate tolerance for data type.

    Parameters
    ----------
    output_data : np.ndarray
        Array from output product.
    compare_data : np.ndarray
        Array from comparison product.

    Returns
    -------
    bool
        True if arrays match within tolerance, False otherwise.
    """
    if np.issubdtype(output_data.dtype, np.floating):
        return np.allclose(
            output_data,
            compare_data,
            rtol=COMPARISON_TOLERANCE,
            atol=COMPARISON_TOLERANCE,
            equal_nan=True,
        )
    return np.array_equal(output_data, compare_data)


def _calculate_difference_stats(
    output_data: np.ndarray, compare_data: np.ndarray
) -> tuple[float, float]:
    """Calculate difference statistics between two arrays.

    Parameters
    ----------
    output_data : np.ndarray
        Array from output product.
    compare_data : np.ndarray
        Array from comparison product.

    Returns
    -------
    tuple[float, float]
        Maximum difference and mean difference.
    """
    diff = np.abs(output_data.astype(np.float64) - compare_data.astype(np.float64))
    return float(np.max(diff)), float(np.mean(diff))


def _log_mismatch(
    message: str,
    output_product: str,
    compare_product: str,
    additional_info: Sequence[str] = (),
) -> None:
    """Log a mismatch between two products.

    Parameters
    ----------
    message : str
        Main message describing the mismatch.
    output_product : str
        Path to output product.
    compare_product : str
        Path to comparison product.
    additional_info : Sequence[str], optional
        Additional information to log.
    """
    log_with_emphasis(
        LOG.interactive,
        message,
        f"output_product: {output_product}",
        f"compare_product: {compare_product}",
        *additional_info,
    )


def _read_geotiff(filepath: Path) -> tuple[np.ndarray, dict]:
    """Read data and profile from a GeoTIFF file.

    Parameters
    ----------
    filepath : Path
        Path to GeoTIFF file.

    Returns
    -------
    tuple[np.ndarray, dict]
        Tuple of (raster data, profile dictionary).
    """
    with rasterio.open(filepath) as src:
        return src.read(), dict(src.profile)


def _write_geotiff(filepath: Path, data: np.ndarray, profile: dict) -> None:
    """Write data to a GeoTIFF file.

    Parameters
    ----------
    filepath : Path
        Path to output GeoTIFF file.
    data : np.ndarray
        Raster data to write.
    profile : dict
        Rasterio profile dictionary.
    """
    with rasterio.open(filepath, "w", **profile) as dst:
        dst.write(data)


def _create_noisy_data(
    data: np.ndarray, scale: float, rng: np.random.Generator
) -> np.ndarray:
    """Create a copy of data with added Gaussian noise.

    Parameters
    ----------
    data : np.ndarray
        Original raster data.
    scale : float
        Standard deviation of the Gaussian noise.
    rng : np.random.Generator
        Random number generator.

    Returns
    -------
    np.ndarray
        Data with added noise.
    """
    noise = rng.normal(scale=scale, size=data.shape)
    return data + noise


def outputs_match(plugin, output_product: str, compare_product: str) -> bool:
    """Compare two GeoTIFF files for equality.

    Parameters
    ----------
    plugin : OutputCheckerPlugin
        The corresponding GeoTIFF output checker (required for API compatibility).
    output_product : str
        Full path to current output product.
    compare_product : str
        Full path to comparison product.

    Returns
    -------
    bool
        True if images match within tolerance, False if they differ.
    """
    output_path = Path(output_product)
    compare_path = Path(compare_product)

    if not all(map(_validate_file_exists, (output_path, compare_path))):
        return False

    try:
        output_data, output_profile = _read_geotiff(output_path)
        compare_data, compare_profile = _read_geotiff(compare_path)

        if not _compare_metadata(output_profile, compare_profile):
            _log_mismatch(
                "BAD GeoTIFFs metadata does NOT match",
                output_product,
                compare_product,
            )
            return False

        if not _compare_arrays(output_data, compare_data):
            max_diff, mean_diff = _calculate_difference_stats(output_data, compare_data)
            _log_mismatch(
                "BAD GeoTIFFs do NOT match",
                output_product,
                compare_product,
                (f"max_diff: {max_diff}", f"mean_diff: {mean_diff}"),
            )
            return False

        log_with_emphasis(LOG.info, "GOOD GeoTIFFs match")
        return True

    except RasterioIOError as e:
        LOG.error("Error reading GeoTIFF files: %s", e)
        return False


def call(plugin, compare_path: str, output_products: list[str]) -> int:
    """Compare correct GeoTIFFs with current output products.

    Compares files produced in the current processing run with the list of
    correct files contained in compare_path.

    Parameters
    ----------
    plugin : OutputCheckerPlugin
        The corresponding GeoTIFF OutputCheckerPlugin.
    compare_path : str
        Path to directory of correct products.
    output_products : list[str]
        List of current output products to compare.

    Returns
    -------
    int
        Binary code: 0 if all comparisons were completed successfully.
    """
    return plugin.compare_outputs(compare_path, output_products)


def _setup_test_directory(test_data_dir: str) -> Path:
    """Create and return the test directory path.

    Parameters
    ----------
    test_data_dir : str
        Base test data directory.

    Returns
    -------
    Path
        Path to the test GeoTIFFs directory.
    """
    savedir = Path(test_data_dir) / "scratch" / "unit_tests" / "test_geotiffs"
    savedir.mkdir(parents=True, exist_ok=True)
    return savedir


def _get_source_tiff_path() -> Path:
    """Get the path to the source test GeoTIFF file.

    Returns
    -------
    Path
        Path to the source GeoTIFF file.
    """
    from importlib.resources import files

    tif_name = "20200405_000000_SH252020_ahi_himawari-8_WV_100kts_100p00_1p0.tif"
    return Path(str(files("geoips") / "../tests/outputs/ahi.tc.WV.geotiff" / tif_name))


def _generate_test_files(
    savedir: Path,
    compare_data: np.ndarray,
    profile: dict,
    rng: np.random.Generator,
) -> tuple[Path, list[Path]]:
    """Generate test GeoTIFF files with various modifications.

    Parameters
    ----------
    savedir : Path
        Directory to save test files.
    compare_data : np.ndarray
        Original comparison data.
    profile : dict
        Rasterio profile for writing files.
    rng : np.random.Generator
        Random number generator.

    Returns
    -------
    tuple[Path, list[Path]]
        Tuple of (compare_file, [matched_file, close_mismatch_file, bad_mismatch_file]).
    """
    file_specs = [
        ("compare.tif", lambda d: d),
        ("matched.tif", lambda d: d),
        (
            "close_mismatch.tif",
            partial(_create_noisy_data, scale=CLOSE_MISMATCH_NOISE_SCALE, rng=rng),
        ),
        (
            "bad_mismatch.tif",
            partial(_create_noisy_data, scale=BAD_MISMATCH_NOISE_SCALE, rng=rng),
        ),
    ]

    generated_paths = []
    for filename, transform in file_specs:
        filepath = savedir / filename
        transformed_data = transform(compare_data)
        _write_geotiff(filepath, transformed_data, profile)
        generated_paths.append(filepath)

    compare_file, *test_files = generated_paths
    return compare_file, test_files


def get_test_files_long(test_data_dir: str) -> tuple[str, list[str]]:
    """Generate a series of GeoTIFF paths for testing comparisons.

    Creates a compare file and test files with varying degrees of modification
    for testing the output checker functionality.

    Parameters
    ----------
    test_data_dir : str
        Base directory for test data.

    Returns
    -------
    tuple[str, list[str]]
        Tuple of (compare_file_path, [matched, close_mismatch, bad_mismatch]).
    """
    import shutil

    rng = np.random.default_rng(RANDOM_SEED)
    savedir = _setup_test_directory(test_data_dir)
    source_path = _get_source_tiff_path()

    temp_compare = savedir / "compare.tif"
    shutil.copy(source_path, temp_compare)

    compare_data, profile = _read_geotiff(temp_compare)
    compare_file, test_files = _generate_test_files(savedir, compare_data, profile, rng)

    return str(compare_file), [str(path) for path in test_files]


def perform_test_comparisons_long(
    plugin, compare_file: str, test_files: list[str]
) -> None:
    """Test comparison of GeoTIFF files with the GeoTIFF Output Checker.

    Parameters
    ----------
    plugin : OutputCheckerPlugin
        The GeoTIFF output checker plugin.
    compare_file : str
        Path to the comparison (baseline) file.
    test_files : list[str]
        List of test file paths [matched, close_mismatch, bad_mismatch].

    Raises
    ------
    AssertionError
        If any comparison result doesn't match expected outcome.
    """
    expected_results = [True, False, False]

    comparison_results = [
        plugin.module.outputs_match(plugin, test_file, compare_file)
        for test_file in test_files
    ]

    for test_file, result, expected in zip(
        test_files, comparison_results, expected_results
    ):
        assert result is expected, (
            f"Unexpected comparison result for {test_file}: "
            f"got {result}, expected {expected}"
        )
